# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the starter.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='Starter',
    description='Models the starter (motor downstream the engine).'
)

dsp.add_data('starter_efficiency', dfl.values.starter_efficiency)
dsp.add_data('delta_time_engine_starter', dfl.values.delta_time_engine_starter)


@sh.add_function(dsp, outputs=['start_demand_function'])
def define_engine_start_demand_function(engine_moment_inertia):
    """
    Define the energy required to start the engine function [kJ].

    :param engine_moment_inertia:
        Engine moment of inertia [kg*m2].
    :type engine_moment_inertia: float

    :return:
        Energy required to start the engine function.
    :rtype: function
    """
    coef = (np.pi / 30) ** 2 / 2000 * engine_moment_inertia

    def _func(start_engine_speed):
        return coef * start_engine_speed ** 2

    return _func


@sh.add_function(dsp, outputs=['starter_powers'])
def calculate_starter_powers(
        start_demand_function, times, on_engine, delta_time_engine_starter,
        engine_speeds_out):
    """
    Calculates starter power [kW].

    :param start_demand_function:
        Energy required to start the engine function.
    :type start_demand_function: function

    :param times:
        Time vector.
    :type times: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param delta_time_engine_starter:
        Time elapsed to turn on the engine with electric starter [s].
    :type delta_time_engine_starter: float

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :return:
        Starter power [kW].
    :rtype: numpy.array
    """
    i = np.where(np.bitwise_xor(on_engine[:-1], on_engine[1:]))[0]
    start = on_engine[i + 1]
    j = np.searchsorted(times, times[i] + delta_time_engine_starter + dfl.EPS)
    e = start_demand_function(engine_speeds_out[i + start.astype(int)])
    e /= (times.take(j, mode='clip') - times[i])
    e[~start] *= -1
    p = np.zeros_like(times, float)
    for i, j, e in zip(i, j, e):
        p[i:j] += e
    return p


@sh.add_function(dsp, outputs=['starter_electric_powers'])
def calculate_starter_electric_powers(starter_powers, starter_efficiency):
    """
    Calculates starter electric power [kW].

    :param starter_powers:
        Starter power [kW].
    :type starter_powers: numpy.array | float

    :param starter_efficiency:
        Starter efficiency [-].
    :type starter_efficiency: float

    :return:
        Starter electric power [kW].
    :rtype: numpy.array | float
    """
    eff = starter_efficiency
    return starter_powers * np.where(starter_powers <= 0, eff, 1 / eff)


@sh.add_function(dsp, outputs=['starter_currents'])
def calculate_starter_currents(
        starter_electric_powers, starter_nominal_voltage):
    """
    Calculates starter currents [A].

    :param starter_electric_powers:
        Starter electric power [kW].
    :type starter_electric_powers: numpy.array | float

    :param starter_nominal_voltage:
        Starter nominal voltage [V].
    :type starter_nominal_voltage: float

    :return:
        Starter currents [A].
    :rtype: numpy.array | float
    """
    from .alternator import calculate_alternator_currents as func
    return func(starter_electric_powers, starter_nominal_voltage)


# noinspection PyMissingOrEmptyDocstring
class StarterModel:
    def __init__(self, start_demand_function, starter_nominal_voltage,
                 starter_efficiency, delta_time_engine_starter):
        self.power_demand = start_demand_function
        self.nominal_voltage = starter_nominal_voltage
        self.efficiency = starter_efficiency
        self.time = delta_time_engine_starter

    def __call__(self, start_engine_speed):
        return calculate_starter_electric_powers(
            self.power_demand(start_engine_speed), self.efficiency
        )


@sh.add_function(dsp, outputs=['starter_model'])
def define_starter_model(
        start_demand_function, starter_nominal_voltage, starter_efficiency,
        delta_time_engine_starter):
    """
    Defines the starter model.

    :param start_demand_function:
        Energy required to start the engine function.
    :type start_demand_function: function

    :param starter_nominal_voltage:
        Starter nominal voltage [V].
    :type starter_nominal_voltage: float

    :param starter_efficiency:
        Starter efficiency [-].
    :type starter_efficiency: float

    :param delta_time_engine_starter:
        Time elapsed to turn on the engine with electric starter [s].
    :type delta_time_engine_starter: float

    :return:
        Starter model.
    :rtype: StarterModel
    """
    return StarterModel(
        start_demand_function, starter_nominal_voltage, starter_efficiency,
        delta_time_engine_starter
    )



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the electric motors of the vehicle.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.electrics.motors

.. autosummary::
    :nosignatures:
    :toctree: motors/

    alternator
    p0
    p1
    p2
    p3
    p4
    planet
    starter
"""

import schedula as sh
from .alternator import dsp as _alternator
from .p0 import dsp as _p0
from .p1 import dsp as _p1
from .p2 import dsp as _p2
from .p3 import dsp as _p3
from .p4 import dsp as _p4
from .planet import dsp as _planet
from .starter import dsp as _starter

dsp = sh.BlueDispatcher(name='Motors', description='Models the vehicle motors.')

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='alternator',
    dsp=_alternator,
    inputs=(
        'alternator_currents', 'alternator_nominal_voltage', 'motive_powers',
        'alternator_electric_powers', 'alternator_efficiency', 'accelerations',
        'service_battery_state_of_charges', 'service_battery_charging_statuses',
        'alternator_current_model', 'on_engine', 'alternator_charging_currents',
        'service_battery_initialization_time', 'times', 'is_hybrid',
    ),
    outputs=(
        'alternator_current_model', 'alternator_currents', 'alternator_powers',
        'alternator_electric_powers'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motor_p0',
    dsp=_p0,
    inputs=(
        'motor_p0_electric_powers', 'motor_p0_maximum_power', 'motor_p0_powers',
        'motor_p0_maximum_torque', 'motor_p0_speed_ratio', 'motor_p0_torques',
        'motor_p0_efficiency', 'motor_p0_rated_speed', 'motor_p0_speeds',
        'engine_speeds_out', 'has_motor_p0', 'times'
    ),
    outputs=(
        'motor_p0_electric_powers', 'motor_p0_maximum_power', 'motor_p0_powers',
        'motor_p0_maximum_torque', 'motor_p0_speed_ratio', 'motor_p0_torques',
        'motor_p0_maximum_powers', 'motor_p0_rated_speed', 'motor_p0_speeds',
        'engine_speeds_out', 'motor_p0_maximum_power_function', 'has_motor_p0',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motor_p1',
    dsp=_p1,
    inputs=(
        'motor_p1_electric_powers', 'motor_p1_maximum_power', 'motor_p1_powers',
        'motor_p1_maximum_torque', 'motor_p1_speed_ratio', 'motor_p1_torques',
        'motor_p1_efficiency', 'motor_p1_rated_speed', 'motor_p1_speeds',
        'engine_speeds_out', 'has_motor_p1', 'times'
    ),
    outputs=(
        'motor_p1_electric_powers', 'motor_p1_maximum_power', 'motor_p1_powers',
        'motor_p1_maximum_torque', 'motor_p1_speed_ratio', 'motor_p1_torques',
        'motor_p1_maximum_powers', 'motor_p1_rated_speed', 'motor_p1_speeds',
        'engine_speeds_out', 'motor_p1_maximum_power_function', 'has_motor_p1',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motor_p2_planetary',
    dsp=_planet,
    inputs=(
        'motor_p2_planetary_powers', 'final_drive_speeds_in', 'planetary_ratio',
        'motor_p2_planetary_maximum_power', 'motor_p2_planetary_maximum_torque',
        'planetary_mean_efficiency', 'planetary_speeds_in', 'engine_speeds_out',
        'motor_p2_planetary_rated_speed', 'motor_p2_planetary_torques', 'times',
        'motor_p2_planetary_electric_powers', 'motor_p2_planetary_speed_ratio',
        'motor_p2_planetary_efficiency', 'motor_p2_planetary_speeds',
        'has_motor_p2_planetary',
    ),
    outputs=(
        'motor_p2_planetary_powers', 'planetary_speeds_in', 'engine_speeds_out',
        'motor_p2_planetary_maximum_power', 'motor_p2_planetary_maximum_torque',
        'motor_p2_planetary_speeds', 'final_drive_speeds_in', 'planetary_ratio',
        'motor_p2_planetary_electric_powers', 'motor_p2_planetary_speed_ratio',
        'motor_p2_planetary_maximum_powers', 'motor_p2_planetary_rated_speed',
        'motor_p2_planetary_maximum_power_function', 'has_motor_p2_planetary',
        'motor_p2_planetary_torques', 'planetary_mean_efficiency',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motor_p2',
    dsp=_p2,
    inputs=(
        'motor_p2_electric_powers', 'motor_p2_maximum_power', 'motor_p2_powers',
        'motor_p2_maximum_torque', 'motor_p2_speed_ratio', 'motor_p2_torques',
        'motor_p2_efficiency', 'motor_p2_rated_speed', 'motor_p2_speeds',
        'gear_box_speeds_in', 'has_motor_p2', 'times'
    ),
    outputs=(
        'motor_p2_electric_powers', 'motor_p2_maximum_power', 'motor_p2_powers',
        'motor_p2_maximum_torque', 'motor_p2_speed_ratio', 'motor_p2_torques',
        'motor_p2_maximum_powers', 'motor_p2_rated_speed', 'motor_p2_speeds',
        'gear_box_speeds_in', 'has_motor_p2',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motor_p3',
    dsp=_p3,
    inputs=(
        'motor_p3_front_electric_powers', 'motor_p3_front_maximum_power',
        'motor_p3_front_powers', 'motor_p3_front_maximum_torque',
        'motor_p3_front_speed_ratio', 'motor_p3_front_torques',
        'motor_p3_front_efficiency', 'motor_p3_front_rated_speed',
        'motor_p3_front_speeds', 'has_motor_p3_front',
        'motor_p3_rear_electric_powers', 'motor_p3_rear_maximum_power',
        'motor_p3_rear_powers', 'motor_p3_rear_maximum_torque',
        'motor_p3_rear_speed_ratio', 'motor_p3_rear_torques',
        'motor_p3_rear_efficiency', 'motor_p3_rear_rated_speed',
        'motor_p3_rear_speeds', 'has_motor_p3_rear', 'final_drive_speeds_in',
        'times',
    ),
    outputs=(
        'motor_p3_front_electric_powers', 'motor_p3_front_maximum_power',
        'motor_p3_front_powers', 'motor_p3_front_maximum_torque',
        'motor_p3_front_speed_ratio', 'motor_p3_front_torques',
        'motor_p3_front_maximum_powers', 'motor_p3_front_rated_speed',
        'motor_p3_front_speeds', 'has_motor_p3_front',
        'motor_p3_rear_electric_powers', 'motor_p3_rear_maximum_power',
        'motor_p3_rear_powers', 'motor_p3_rear_maximum_torque',
        'motor_p3_rear_speed_ratio', 'motor_p3_rear_torques',
        'motor_p3_rear_maximum_powers', 'motor_p3_rear_rated_speed',
        'motor_p3_rear_speeds', 'has_motor_p3_rear', 'final_drive_speeds_in'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motor_p4',
    dsp=_p4,
    inputs=(
        'motor_p4_front_electric_powers', 'motor_p4_front_maximum_power',
        'motor_p4_front_powers', 'motor_p4_front_maximum_torque',
        'motor_p4_front_speed_ratio', 'motor_p4_front_torques',
        'motor_p4_front_efficiency', 'motor_p4_front_rated_speed',
        'motor_p4_front_speeds', 'has_motor_p4_front',
        'motor_p4_rear_electric_powers', 'motor_p4_rear_maximum_power',
        'motor_p4_rear_powers', 'motor_p4_rear_maximum_torque',
        'motor_p4_rear_speed_ratio', 'motor_p4_rear_torques',
        'motor_p4_rear_efficiency', 'motor_p4_rear_rated_speed',
        'motor_p4_rear_speeds', 'has_motor_p4_rear', 'wheel_speeds', 'times',
    ),
    outputs=(
        'motor_p4_front_electric_powers', 'motor_p4_front_maximum_power',
        'motor_p4_front_powers', 'motor_p4_front_maximum_torque',
        'motor_p4_front_speed_ratio', 'motor_p4_front_torques',
        'motor_p4_front_maximum_powers', 'motor_p4_front_rated_speed',
        'motor_p4_front_speeds', 'has_motor_p4_front',
        'motor_p4_rear_electric_powers', 'motor_p4_rear_maximum_power',
        'motor_p4_rear_powers', 'motor_p4_rear_maximum_torque',
        'motor_p4_rear_speed_ratio', 'motor_p4_rear_torques',
        'motor_p4_rear_maximum_powers', 'motor_p4_rear_rated_speed',
        'motor_p4_rear_speeds', 'has_motor_p3_rear', 'wheel_speeds'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='starter',
    dsp=_starter,
    inputs=(
        'delta_time_engine_starter', 'times', 'on_engine', 'starter_efficiency',
        'engine_moment_inertia', 'starter_nominal_voltage', 'engine_speeds_out',
    ),
    outputs=(
        'starter_electric_powers', 'starter_powers', 'starter_currents',
        'starter_model', 'delta_time_engine_starter'
    )
)


@sh.add_function(dsp, outputs=['is_hybrid'])
def identify_is_hybrid(
        has_motor_p0, has_motor_p1, has_motor_p2, has_motor_p2_planetary,
        has_motor_p3_front, has_motor_p3_rear, has_motor_p4_front,
        has_motor_p4_rear):
    """
    Identifies if the the vehicle is hybrid.

    :param has_motor_p0:
        Has the vehicle a motor in P0?
    :type has_motor_p0: bool

    :param has_motor_p1:
        Has the vehicle a motor in P1?
    :type has_motor_p1: bool

    :param has_motor_p2:
        Has the vehicle a motor in P2?
    :type has_motor_p2: bool

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :param has_motor_p3_front:
        Has the vehicle a motor in P3 front?
    :type has_motor_p3_front: bool

    :param has_motor_p3_rear:
        Has the vehicle a motor in P3 rear?
    :type has_motor_p3_rear: bool

    :param has_motor_p4_front:
        Has the vehicle a motor in P4 front?
    :type has_motor_p4_front: bool

    :param has_motor_p4_rear:
        Has the vehicle a motor in P4 rear?
    :type has_motor_p4_rear: bool

    :return:
        Is the vehicle hybrid?
    :rtype: bool
    """
    b = has_motor_p0, has_motor_p1, has_motor_p2, has_motor_p2_planetary,
    b += has_motor_p3_front, has_motor_p3_rear, has_motor_p4_front
    b += has_motor_p4_rear,
    return any(b)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the vehicle batteries.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.electrics.batteries

.. autosummary::
    :nosignatures:
    :toctree: batteries/

    dcdc
    drive
    service
"""

import schedula as sh
from .service import dsp as _service
from .drive import dsp as _drive
from .dcdc import dsp as _dcdc

dsp = sh.BlueDispatcher(
    name='Batteries', description='Models the vehicle batteries.'
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='service_battery',
    dsp=_service,
    inputs=(
        'initial_service_battery_state_of_charge', 'alternator_electric_powers',
        'service_battery_start_window_width', 'service_battery_nominal_voltage',
        'service_battery_state_of_charge_balance', 'on_engine', 'accelerations',
        'cycle_type', 'service_battery_electric_powers', 'service_battery_load',
        'service_battery_electric_powers_supply_threshold', 'engine_powers_out',
        'service_battery_state_of_charge_balance_window', 'is_hybrid', 'times',
        'alternator_current_model', 'has_energy_recuperation', 'motive_powers',
        'service_battery_capacity', 'starter_electric_powers', 'engine_starts',
        'service_battery_initialization_time', 'service_battery_status_model',
        'service_battery_electric_powers_supply', 'service_battery_currents',
        'dcdc_converter_electric_powers', 'service_battery_state_of_charges',
        'service_battery_capacity_kwh', 'service_battery_loads',
        'dcdc_current_model',
    ),
    outputs=(
        'service_battery_charging_statuses', 'service_battery_state_of_charges',
        'service_battery_delta_state_of_charge', 'service_battery_capacity_kwh',
        'initial_service_battery_state_of_charge', 'service_battery_currents',
        'service_battery_state_of_charge_balance', 'service_battery_capacity',
        'service_battery_initialization_time', 'service_battery_status_model',
        'service_battery_electric_powers', 'service_battery_loads',
        'service_battery_electric_powers_supply_threshold',
        'service_battery_state_of_charge_balance_window',
        'service_battery_model', 'service_battery_load',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='drive_battery',
    dsp=_drive,
    inputs=(
        'dcdc_converter_electric_powers_demand', 'drive_battery_n_series_cells',
        'motor_p1_electric_powers', 'drive_battery_loads', 'drive_battery_load',
        'motor_p2_planetary_electric_powers', 'electrical_hybridization_degree',
        'drive_battery_currents', 'drive_battery_capacity', 'drive_battery_r0',
        'drive_battery_n_parallel_cells', 'dcdc_converter_efficiency', 'times',
        'drive_battery_voltages', 'service_battery_model', 'drive_battery_ocv',
        'initial_drive_battery_state_of_charge', 'motor_p0_electric_powers',
        'drive_battery_electric_powers', 'drive_battery_state_of_charges',
        'motor_p3_front_electric_powers', 'motor_p3_rear_electric_powers',
        'motor_p4_front_electric_powers', 'motor_p4_rear_electric_powers',
        'drive_battery_nominal_voltage', 'motor_p2_electric_powers',
        'drive_battery_capacity_kwh', 'drive_battery_technology',
        'drive_battery_n_cells',
    ),
    outputs=(
        'drive_battery_n_series_cells', 'drive_battery_ocv', 'drive_battery_r0',
        'drive_battery_delta_state_of_charge', 'drive_battery_n_parallel_cells',
        'drive_battery_voltages', 'drive_battery_model', 'drive_battery_load',
        'initial_drive_battery_state_of_charge', 'drive_battery_capacity',
        'drive_battery_state_of_charges', 'drive_battery_nominal_voltage',
        'drive_battery_electric_powers', 'motors_electric_powers',
        'drive_battery_capacity_kwh', 'drive_battery_currents',
        'drive_battery_n_cells', 'drive_battery_loads',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='dcdc_converter',
    dsp=_dcdc,
    inputs=(
        'service_battery_state_of_charges', 'service_battery_charging_statuses',
        'dcdc_converter_electric_powers', 'service_battery_initialization_time',
        'service_battery_nominal_voltage', 'dcdc_converter_efficiency', 'times',
        'dcdc_converter_electric_powers_demand', 'dcdc_converter_currents',
        'dcdc_charging_currents', 'dcdc_current_model', 'is_hybrid',
    ),
    outputs=(
        'dcdc_converter_electric_powers_demand', 'dcdc_current_model',
        'dcdc_converter_electric_powers', 'dcdc_converter_currents'
    )
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to define a sub-model from a dsp.
"""

_cache_sub_model = {}


def define_sub_model(dsp, inputs, outputs, models):
    """
    Defines a sub-model from a dsp.

    :param dsp:
        Original model.
    :type dsp: schedula.Dispatcher

    :param inputs:
        Data inputs.
    :type inputs: list | tuple

    :param outputs:
        Data outputs.
    :type outputs: list | tuple

    :param models:
        Data models.
    :type models: list | tuple

    :return:
        A sub-model.
    :rtype: schedula.Dispatcher
    """
    sub_id = dsp.name, tuple(inputs), tuple(outputs), tuple(models)
    if sub_id in _cache_sub_model:
        return _cache_sub_model[sub_id].copy()
    assert not set(models).difference(dsp.nodes), models
    sub = dsp.shrink_dsp(set(inputs or []).union(models), outputs)
    assert set(sub.nodes).issuperset(set(inputs).union(outputs)), models
    _cache_sub_model[sub_id] = sub
    return sub


def _accuracy_score(*args, **kwargs):
    from sklearn.metrics import accuracy_score as func
    return func(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the final drive.
"""

import logging
import collections
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

log = logging.getLogger(__name__)

dsp = sh.BlueDispatcher(
    name='Final drive', description='Models the final drive.'
)
dsp.add_data('final_drive_ratio', dfl.values.final_drive_ratio)


@sh.add_function(dsp, inputs_kwargs=True, outputs=['final_drive_ratios'])
def calculate_final_drive_ratios(final_drive_ratio, n_gears=1):
    """
    Defines final drive ratios for each gear [-].

    :param final_drive_ratio:
        Final drive ratio [-].
    :type final_drive_ratio: float

    :param n_gears:
        Number of gears [-].
    :type n_gears: int, optional

    :return:
        Final drive ratios [-].
    :rtype: dict
    """
    d = collections.defaultdict(lambda: final_drive_ratio)
    d.update(dict.fromkeys(range(0, int(n_gears + 1)), final_drive_ratio))
    return d


# noinspection PyUnusedLocal,PyMissingOrEmptyDocstring
def is_not_manual_or_automatic(gear_box_type, *args):
    return gear_box_type not in ('manual', 'automatic')


dsp.add_function(
    function=sh.add_args(calculate_final_drive_ratios),
    inputs=['gear_box_type', 'final_drive_ratio'],
    outputs=['final_drive_ratios'],
    input_domain=is_not_manual_or_automatic
)


@sh.add_function(dsp, outputs=['final_drive_ratio_vector'])
def calculate_final_drive_ratio_vector(final_drive_ratios, gears):
    """
    Calculates the final drive ratio vector [-].

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :return:
        Final drive ratio vector [-].
    :rtype: numpy.array
    """
    fdr0 = final_drive_ratios[min(final_drive_ratios)]
    d = collections.defaultdict(lambda: fdr0)
    d.update(final_drive_ratios)
    return np.vectorize(lambda k: d[k])(gears)


@sh.add_function(dsp, outputs=['final_drive_speeds_in'])
def calculate_final_drive_speeds_in(
        final_drive_speeds_out, final_drive_ratio_vector):
    """
    Calculates final drive speed [RPM].

    :param final_drive_speeds_out:
        Rotating speed of the wheel [RPM].
    :type final_drive_speeds_out: numpy.array | float

    :param final_drive_ratio_vector:
        Final drive ratio vector [-].
    :type final_drive_ratio_vector: numpy.array | float

    :return:
        Final drive speed in [RPM].
    :rtype: numpy.array | float
    """
    return final_drive_speeds_out * final_drive_ratio_vector


dsp.add_data('wheel_drive', dfl.values.wheel_drive)


@sh.add_function(dsp, outputs=['n_wheel_drive'])
def define_n_wheel_drive(wheel_drive):
    """
    Defines the default number of wheel drive [-].

    :param wheel_drive:
        Wheel drive (i.e., front, rear, front+rear).
    :type wheel_drive: str

    :return:
        Number of wheel drive [-].
    :rtype: int
    """
    if 'front+rear' == wheel_drive:
        return 4
    return 2


@sh.add_function(dsp, outputs=['final_drive_efficiency'])
def default_final_drive_efficiency(n_wheel_drive):
    """
    Returns the default final drive efficiency [-].

    :param n_wheel_drive:
        Number of wheel drive [-].
    :type n_wheel_drive: int

    :return:
        Final drive efficiency [-].
    :rtype: float
    """
    from asteval import Interpreter as Interp
    formula = dfl.functions.default_final_drive_efficiency.formula
    return Interp(dict(n_wheel_drive=n_wheel_drive)).eval(formula)


@sh.add_function(dsp, outputs=['final_drive_powers_in'])
def calculate_final_drive_powers_in(
        final_drive_powers_out, final_drive_efficiency):
    """
    Calculate final drive power in [kW].

    :param final_drive_powers_out:
        Final drive power out [kW].
    :type final_drive_powers_out: numpy.array | float

    :param final_drive_efficiency:
        Final drive efficiency [-].
    :type final_drive_efficiency: float

    :return:
        Final drive power in [kW].
    :rtype: numpy.array | float
    """
    eff, p = final_drive_efficiency, final_drive_powers_out
    return np.where(p < 0, eff, 1 / eff) * p


@sh.add_function(dsp, outputs=['final_drive_powers_out'])
def calculate_final_drive_powers_out(
        final_drive_powers_in, final_drive_efficiency):
    """
    Calculate final drive power out [kW].

    :param final_drive_powers_in:
        Final drive power in [kW].
    :type final_drive_powers_in: numpy.array | float

    :param final_drive_efficiency:
        Final drive efficiency [-].
    :type final_drive_efficiency: float

    :return:
        Final drive power out [kW].
    :rtype: numpy.array | float
    """
    return calculate_final_drive_powers_in(
        final_drive_powers_in, 1 / final_drive_efficiency
    )


@sh.add_function(dsp, outputs=['final_drive_torques_in'])
def calculate_final_drive_torques_in(
        final_drive_powers_in, final_drive_speeds_in):
    """
    Calculate final drive power in [kW].

    :param final_drive_powers_in:
        Final drive power in [kW].
    :type final_drive_powers_in: numpy.array | float

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array | float

    :return:
        Final drive torque in [N*m].
    :rtype: numpy.array | float
    """
    from .wheels import calculate_wheel_torques as func
    return func(final_drive_powers_in, final_drive_speeds_in)


dsp.add_function(
    function=sh.bypass,
    inputs=['final_drive_efficiency'],
    outputs=['final_drive_mean_efficiency']
)


# noinspection PyUnusedLocal
def domain_final_drive_torque_losses_v1(n_dyno_axes, n_wheel_drive, *args):
    """
    Check the validity of number of wheel drive respect to the dyno axes
    assuming 2 wheels per axes.

    :param n_dyno_axes:
        Number of dyno axes [-].
    :type n_dyno_axes: int

    :param n_wheel_drive:
        Number of wheel drive [-].
    :type n_wheel_drive: int

    :return:
        True and log a waring if `n_wheel_drive` does not respect the domain.
    :rtype: bool
    """

    if n_dyno_axes < n_wheel_drive / 2:
        msg = 'WARNING: n_dyno_axes(%d) < n_wheel_drive(%d) / 2!'
        log.warning(msg, n_dyno_axes, n_wheel_drive)
    return True



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It provides CO2MPAS architecture model `dsp`.

Sub-Modules:

.. currentmodule:: co2mpas.core.model

.. autosummary::
    :nosignatures:
    :toctree: model/

    physical
    selector
"""
import numpy as np
import schedula as sh
from .physical import dsp as _physical
from .selector import dsp as _selector, calibration_cycles, prediction_cycles

dsp = sh.BlueDispatcher(
    name='CO2MPAS model',
    description='Calibrates the models with WLTP data and predicts NEDC cycle.'
)

_physical = sh.SubDispatch(_physical)

dsp.add_data(
    data_id='input.calibration.wltp_h',
    description='User input data of WLTP-H calibration stage.',
    default_value=None,
    filters=[lambda x: {} if x is None else x]
)
dsp.add_data(
    data_id='input.calibration.wltp_l',
    description='User input data of WLTP-L calibration stage.',
    default_value=None,
    filters=[lambda x: {} if x is None else x]
)
dsp.add_data(
    data_id='input.calibration.wltp_m',
    description='User input data of WLTP-M calibration stage.',
    default_value=None,
    filters=[lambda x: {} if x is None else x]
)
dsp.add_data(
    data_id='output.calibration.wltp_h',
    description='Output data of WLTP-H calibration stage.'
)
dsp.add_data(
    data_id='output.calibration.wltp_l',
    description='Output data of WLTP-L calibration stage.'
)
dsp.add_data(
    data_id='output.calibration.wltp_m',
    description='Output data of WLTP-M calibration stage.'
)
dsp.add_function(
    function_id='calibrate_with_wltp_h',
    function=_physical,
    inputs=['input.calibration.wltp_h'],
    outputs=['output.calibration.wltp_h'],
    description='Wraps all functions needed to calibrate the models to '
                'predict CO2 emissions.'
)

dsp.add_function(
    function_id='calibrate_with_wltp_l',
    function=_physical,
    inputs=['input.calibration.wltp_l'],
    outputs=['output.calibration.wltp_l'],
    description='Wraps all functions needed to calibrate the models to '
                'predict CO2 emissions.'
)
dsp.add_function(
    function_id='calibrate_with_wltp_m',
    function=_physical,
    inputs=['input.calibration.wltp_m'],
    outputs=['output.calibration.wltp_m'],
    description='Wraps all functions needed to calibrate the models to '
                'predict CO2 emissions.'
)

dsp.add_data(
    data_id='input.prediction.wltp_h', default_value=None,
    description='User input data of WLTP-H prediction stage.',
    filters=[lambda x: {} if x is None else x]
)
dsp.add_data(
    data_id='input.prediction.wltp_l', default_value=None,
    description='User input data of WLTP-L prediction stage.',
    filters=[lambda x: {} if x is None else x]
)
dsp.add_data(
    data_id='input.prediction.wltp_m', default_value=None,
    description='User input data of WLTP-M prediction stage.',
    filters=[lambda x: {} if x is None else x]
)
dsp.add_data(
    data_id='data.prediction.models_wltp_h',
    description='Calibrated models for WLTP-H prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.models_wltp_l',
    description='Calibrated models for WLTP-L prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.models_wltp_m',
    description='Calibrated models for WLTP-M prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.wltp_h',
    description='Input data of WLTP-H prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.wltp_l',
    description='Input data of WLTP-L prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.wltp_m',
    description='Input data of WLTP-M prediction stage.'
)


@sh.add_function(
    dsp, inputs=['output.calibration.wltp_h', 'data.prediction.models_wltp_h',
                 'input.prediction.wltp_h'], outputs=['data.prediction.wltp_h']
)
@sh.add_function(
    dsp, inputs=['output.calibration.wltp_l', 'data.prediction.models_wltp_l',
                 'input.prediction.wltp_l'], outputs=['data.prediction.wltp_l']
)
@sh.add_function(
    dsp, inputs=['output.calibration.wltp_m', 'data.prediction.models_wltp_m',
                 'input.prediction.wltp_m'], outputs=['data.prediction.wltp_m']
)
def select_prediction_data(calibration_data, models_data, user_data):
    """
    Selects the data required to predict the CO2 emissions with CO2MPAS model.

    :param calibration_data:
        Output data of calibration stage
    :type calibration_data: dict

    :param models_data:
        Calibrated models for prediction stage.
    :type models_data: dict

    :param user_data:
        User input data of prediction stage.
    :type user_data: dict

    :return:
        Data required to predict the CO2 emissions with CO2MPAS model.
    :rtype: dict
    """
    from co2mpas.defaults import dfl
    kw = dict(calibration=calibration_data, models=models_data, user=user_data)
    data, xp = [], None
    for k, v in dfl.functions.select_prediction_data.prediction_data[::-1]:
        d = kw[k]
        if v != 'all':
            d = sh.selector(v, d, allow_miss=True)
        if 'times' in d:
            if xp is None:
                xp = d['times']
            else:
                x, inter = d['times'], np.interp
                d = {k: inter(x, xp, fp) for k, fp in d.items() if k != 'times'}
        data.append(d)
    return sh.combine_dicts(*data[::-1])


dsp.add_data(
    data_id='input.prediction.models', default_value={},
    description='User input models for prediction stages.'
)
dsp.add_data(
    data_id='enable_selector', default_value=False,
    description='Enable the selection of the best model to predict '
                'H/L/M cycles.'
)
dsp.add_data(
    data_id='data.calibration.model_scores',
    description='Scores of calibrated models.'
)

_prediction_models = tuple(map('models_{}'.format, prediction_cycles))
dsp.add_function(
    function=sh.SubDispatchPipe(
        _selector,
        function_id='extract_calibrated_models',
        inputs=('enable_selector', 'default_models',) + calibration_cycles,
        outputs=('selections',) + _prediction_models
    ),
    inputs=[
        'enable_selector', 'input.prediction.models',
        'output.calibration.wltp_h', 'output.calibration.wltp_l',
        'output.calibration.wltp_m'
    ],
    outputs=['data.calibration.model_scores'] + [
        'data.prediction.%s' % k for k in _prediction_models
    ]
)

dsp.add_data(
    data_id='output.prediction.wltp_h',
    description='Output data of WLTP-H prediction stage.'
)
dsp.add_data(
    data_id='output.prediction.wltp_l',
    description='Output data of WLTP-L prediction stage.'
)
dsp.add_data(
    data_id='output.prediction.wltp_m',
    description='Output data of WLTP-M prediction stage.'
)
dsp.add_function(
    function_id='predict_wltp_h',
    function=_physical,
    inputs=['data.prediction.wltp_h'],
    outputs=['output.prediction.wltp_h'],
    description='Wraps all functions needed to predict CO2 emissions.'
)

dsp.add_function(
    function_id='predict_wltp_l',
    function=_physical,
    inputs=['data.prediction.wltp_l'],
    outputs=['output.prediction.wltp_l'],
    description='Wraps all functions needed to predict CO2 emissions.'
)

dsp.add_function(
    function_id='predict_wltp_m',
    function=_physical,
    inputs=['data.prediction.wltp_m'],
    outputs=['output.prediction.wltp_m'],
    description='Wraps all functions needed to predict CO2 emissions.'
)

dsp.add_data(
    data_id='input.prediction.nedc_h',
    description='User input data of NEDC-H prediction stage.'
)
dsp.add_data(
    data_id='input.prediction.nedc_l',
    description='User input data of NEDC-L prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.models_nedc_h',
    description='Calibrated models for NEDC-H prediction stage.'
)
dsp.add_data(
    data_id='data.prediction.models_nedc_l',
    description='Calibrated models for NEDC-L prediction stage.'
)
dsp.add_data(
    data_id='output.prediction.nedc_h',
    description='Output data of NEDC-H prediction stage.'
)
dsp.add_data(
    data_id='output.prediction.nedc_l',
    description='Output data of NEDC-L prediction stage.'
)
dsp.add_function(
    function_id='predict_nedc_h',
    function=_physical,
    inputs=['data.prediction.models_nedc_h', 'input.prediction.nedc_h'],
    outputs=['output.prediction.nedc_h'],
    description='Wraps all functions needed to predict CO2 emissions.'
)

dsp.add_function(
    function_id='predict_nedc_l',
    function=_physical,
    inputs=['data.prediction.models_nedc_l', 'input.prediction.nedc_l'],
    outputs=['output.prediction.nedc_l'],
    description='Wraps all functions needed to predict CO2 emissions.'
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the electrics of the vehicle.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.electrics

.. autosummary::
    :nosignatures:
    :toctree: electrics/

    motors
    batteries
"""
import numpy as np
import schedula as sh
from .motors import dsp as _motors
from .batteries import dsp as _batteries

dsp = sh.BlueDispatcher(
    name='Electrics', description='Models the vehicle electrics.'
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='motors',
    dsp=_motors,
    inputs=(
        'motor_p0_efficiency', 'motor_p0_maximum_power', 'motor_p0_speed_ratio',
        'motor_p0_electric_powers', 'motor_p0_torques', 'motor_p0_rated_speed',
        'motor_p0_powers', 'motor_p0_maximum_torque', 'motor_p0_speeds',
        'motor_p1_efficiency', 'motor_p1_maximum_power', 'motor_p1_speed_ratio',
        'motor_p1_electric_powers', 'motor_p1_torques', 'motor_p1_rated_speed',
        'motor_p1_powers', 'motor_p1_maximum_torque', 'motor_p1_speeds',
        'motor_p2_planetary_efficiency', 'motor_p2_planetary_maximum_power',
        'motor_p2_planetary_speed_ratio', 'planetary_ratio',
        'motor_p2_planetary_electric_powers', 'motor_p2_planetary_torques',
        'motor_p2_planetary_powers', 'motor_p2_planetary_maximum_torque',
        'motor_p2_planetary_rated_speed', 'motor_p2_planetary_speeds',
        'motor_p2_efficiency', 'motor_p2_maximum_power', 'motor_p2_speed_ratio',
        'motor_p2_electric_powers', 'motor_p2_torques', 'motor_p2_rated_speed',
        'motor_p2_powers', 'motor_p2_maximum_torque', 'motor_p2_speeds',
        'motor_p3_front_efficiency', 'motor_p3_front_maximum_power',
        'motor_p3_front_speed_ratio', 'motor_p3_front_electric_powers',
        'motor_p3_front_torques', 'motor_p3_front_rated_speed',
        'motor_p3_front_powers', 'motor_p3_front_maximum_torque',
        'motor_p3_front_speeds',
        'motor_p3_rear_efficiency', 'motor_p3_rear_maximum_power',
        'motor_p3_rear_speed_ratio', 'motor_p3_rear_electric_powers',
        'motor_p3_rear_torques', 'motor_p3_rear_rated_speed',
        'motor_p3_rear_powers', 'motor_p3_rear_maximum_torque',
        'motor_p3_rear_speeds',
        'motor_p4_front_efficiency', 'motor_p4_front_maximum_power',
        'motor_p4_front_speed_ratio', 'motor_p4_front_electric_powers',
        'motor_p4_front_torques', 'motor_p4_front_rated_speed',
        'motor_p4_front_powers', 'motor_p4_front_maximum_torque',
        'motor_p4_front_speeds',
        'motor_p4_rear_efficiency', 'motor_p4_rear_maximum_power',
        'motor_p4_rear_speed_ratio', 'motor_p4_rear_electric_powers',
        'motor_p4_rear_torques', 'motor_p4_rear_rated_speed',
        'motor_p4_rear_powers', 'motor_p4_rear_maximum_torque',
        'motor_p4_rear_speeds',
        'has_motor_p0', 'has_motor_p1', 'has_motor_p2', 'has_motor_p3_front',
        'has_motor_p3_rear', 'has_motor_p4_front', 'has_motor_p4_rear',
        'service_battery_initialization_time', 'engine_starts', 'accelerations',
        'starter_efficiency', 'delta_time_engine_starter', 'gear_box_speeds_in',
        'service_battery_charging_statuses', 'alternator_currents', 'on_engine',
        'starter_nominal_voltage', 'engine_speeds_out', 'alternator_efficiency',
        'service_battery_state_of_charges', 'alternator_current_model', 'times',
        'alternator_charging_currents', 'engine_moment_inertia', 'wheel_speeds',
        'alternator_electric_powers', 'final_drive_speeds_in', 'motive_powers',
        'alternator_nominal_voltage', 'planetary_mean_efficiency',
        'planetary_speeds_in',
    ),
    outputs=(
        'motor_p0_electric_powers', 'motor_p0_maximum_power', 'motor_p0_powers',
        'motor_p0_maximum_torque', 'motor_p0_speed_ratio', 'motor_p0_torques',
        'motor_p0_maximum_powers', 'motor_p0_rated_speed', 'motor_p0_speeds',
        'motor_p0_maximum_power_function', 'motor_p1_maximum_power_function',
        'motor_p1_electric_powers', 'motor_p1_maximum_power', 'motor_p1_powers',
        'motor_p1_maximum_torque', 'motor_p1_speed_ratio', 'motor_p1_torques',
        'motor_p1_maximum_powers', 'motor_p1_rated_speed', 'motor_p1_speeds',
        'motor_p2_planetary_maximum_power_function',
        'motor_p2_planetary_electric_powers', 'planetary_ratio',
        'motor_p2_planetary_maximum_power', 'motor_p2_planetary_powers',
        'motor_p2_planetary_maximum_torque', 'motor_p2_planetary_speed_ratio',
        'motor_p2_planetary_maximum_powers', 'motor_p2_planetary_rated_speed',
        'motor_p2_planetary_torques', 'motor_p2_planetary_speeds',
        'motor_p2_electric_powers', 'motor_p2_maximum_power', 'motor_p2_powers',
        'motor_p2_maximum_torque', 'motor_p2_speed_ratio', 'motor_p2_torques',
        'motor_p2_maximum_powers', 'motor_p2_rated_speed', 'motor_p2_speeds',
        'motor_p3_front_electric_powers', 'motor_p3_front_maximum_power',
        'motor_p3_front_maximum_powers', 'motor_p3_front_rated_speed',
        'motor_p3_front_powers', 'motor_p3_front_maximum_torque',
        'motor_p3_front_speed_ratio', 'motor_p3_front_torques',
        'motor_p3_front_speeds',
        'motor_p3_rear_electric_powers', 'motor_p3_rear_maximum_power',
        'motor_p3_rear_maximum_powers', 'motor_p3_rear_rated_speed',
        'motor_p3_rear_powers', 'motor_p3_rear_maximum_torque',
        'motor_p3_rear_speed_ratio', 'motor_p3_rear_torques',
        'motor_p3_rear_speeds',
        'motor_p4_front_electric_powers', 'motor_p4_front_maximum_power',
        'motor_p4_front_powers', 'motor_p4_front_maximum_torque',
        'motor_p4_front_speed_ratio', 'motor_p4_front_torques',
        'motor_p4_front_maximum_powers', 'motor_p4_front_rated_speed',
        'motor_p4_front_speeds',
        'motor_p4_rear_electric_powers', 'motor_p4_rear_maximum_power',
        'motor_p4_rear_powers', 'motor_p4_rear_maximum_torque',
        'motor_p4_rear_speed_ratio', 'motor_p4_rear_torques',
        'motor_p4_rear_maximum_powers', 'motor_p4_rear_rated_speed',
        'motor_p4_rear_speeds',
        'starter_electric_powers', 'final_drive_speeds_in', 'alternator_powers',
        'alternator_electric_powers', 'engine_speeds_out', 'gear_box_speeds_in',
        'delta_time_engine_starter', 'alternator_currents', 'starter_currents',
        'alternator_current_model', 'starter_powers', 'starter_model',
        'wheel_speeds', 'planetary_speeds_in', 'has_motor_p0', 'has_motor_p1',
        'has_motor_p2_planetary', 'has_motor_p2', 'has_motor_p3_front',
        'has_motor_p3_rear', 'has_motor_p4_front', 'has_motor_p4_rear',
        'is_hybrid', 'planetary_mean_efficiency',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='batteries',
    dsp=_batteries,
    inputs=(
        'drive_battery_ocv', 'drive_battery_r0', 'drive_battery_n_series_cells',
        'initial_service_battery_state_of_charge', 'alternator_electric_powers',
        'service_battery_start_window_width', 'service_battery_nominal_voltage',
        'cycle_type', 'service_battery_electric_powers', 'service_battery_load',
        'initial_drive_battery_state_of_charge', 'service_battery_status_model',
        'motor_p2_electric_powers', 'service_battery_capacity', 'accelerations',
        'service_battery_loads', 'dcdc_charging_currents', 'dcdc_current_model',
        'service_battery_electric_powers_supply_threshold', 'engine_powers_out',
        'motor_p0_electric_powers', 'alternator_current_model', 'motive_powers',
        'service_battery_state_of_charge_balance_window', 'drive_battery_loads',
        'drive_battery_state_of_charges', 'drive_battery_currents', 'on_engine',
        'service_battery_state_of_charges', 'motor_p1_electric_powers', 'times',
        'service_battery_initialization_time', 'dcdc_converter_electric_powers',
        'has_energy_recuperation', 'drive_battery_voltages', 'starter_currents',
        'dcdc_converter_currents', 'drive_battery_electric_powers', 'is_hybrid',
        'service_battery_state_of_charge_balance', 'dcdc_converter_efficiency',
        'motor_p2_planetary_electric_powers', 'drive_battery_nominal_voltage',
        'drive_battery_technology', 'drive_battery_capacity', 'engine_starts',
        'drive_battery_n_parallel_cells', 'electrical_hybridization_degree',
        'dcdc_converter_electric_powers_demand', 'starter_electric_powers',
        'motor_p3_front_electric_powers', 'motor_p3_rear_electric_powers',
        'motor_p4_front_electric_powers', 'motor_p4_rear_electric_powers',
        'service_battery_capacity_kwh', 'drive_battery_capacity_kwh',
        'service_battery_currents', 'drive_battery_n_cells',
        'drive_battery_load',
    ),
    outputs=(
        'service_battery_electric_powers_supply_threshold', 'drive_battery_ocv',
        'service_battery_state_of_charge_balance_window', 'drive_battery_model',
        'drive_battery_state_of_charges', 'drive_battery_delta_state_of_charge',
        'initial_drive_battery_state_of_charge', 'drive_battery_n_series_cells',
        'dcdc_converter_currents', 'drive_battery_capacity', 'drive_battery_r0',
        'service_battery_delta_state_of_charge', 'service_battery_status_model',
        'service_battery_state_of_charges', 'service_battery_charging_statuses',
        'drive_battery_n_cells', 'service_battery_loads', 'drive_battery_loads',
        'drive_battery_voltages', 'service_battery_model', 'drive_battery_load',
        'dcdc_converter_electric_powers_demand', 'service_battery_capacity_kwh',
        'service_battery_initialization_time', 'drive_battery_electric_powers',
        'service_battery_state_of_charge_balance', 'dcdc_converter_efficiency',
        'motors_electric_powers', 'starter_currents', 'drive_battery_currents',
        'initial_service_battery_state_of_charge', 'service_battery_capacity',
        'service_battery_electric_powers', 'drive_battery_n_parallel_cells',
        'drive_battery_nominal_voltage', 'drive_battery_capacity_kwh',
        'service_battery_currents', 'service_battery_load',
        'dcdc_current_model',
    )
)

dsp.add_function(
    function=sh.bypass,
    inputs=['service_battery_nominal_voltage'],
    outputs=['alternator_nominal_voltage'],
)

dsp.add_function(
    function=sh.bypass,
    inputs=['alternator_nominal_voltage'],
    outputs=['starter_nominal_voltage'],
)

dsp.add_function(
    function=sh.bypass,
    inputs=['starter_nominal_voltage'],
    outputs=['service_battery_nominal_voltage'],
)


@sh.add_function(dsp, outputs=[
    'service_battery_state_of_charges', 'service_battery_charging_statuses',
    'dcdc_converter_currents', 'alternator_currents'
])
def predict_service_battery_flows(
        service_battery_model, times, motive_powers, accelerations, on_engine,
        starter_currents):
    """
    Predict the service battery currents flows.

    :param service_battery_model:
         Service battery model.
    :type service_battery_model: ServiceBatteryModel

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param accelerations:
        Acceleration [m/s2].
    :type accelerations: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param starter_currents:
        Starter currents [A].
    :type starter_currents: numpy.array

    :return:
        - State of charge of the service battery [%].
        - Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
          3: Initialization) [-].
        - DC/DC converter currents [A].
        - Alternator currents [A].
    :rtype: numpy.array
    """
    service_battery_model.reset()
    it = zip(times, motive_powers, accelerations, on_engine, starter_currents)
    return np.array([service_battery_model(*a) for a in it], float).T



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to compare/select the calibrated models.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.selector.models

.. autosummary::
    :nosignatures:
    :toctree: models/

    _core
    at_model
    after_treatment_model
    clutch_torque_converter_model
    co2_params
    control_model
    electrics_model
    engine_coolant_temperature_model
    engine_speed_model
"""
import logging
import functools
import numpy as np
import schedula as sh
from collections.abc import Iterable

log = logging.getLogger(__name__)
calibration_cycles = 'wltp_h', 'wltp_l', 'wltp_m'
prediction_cycles = 'nedc_h', 'nedc_l', 'wltp_h', 'wltp_l', 'wltp_m'

_select_data = functools.partial(sh.selector, allow_miss=True)
_map_list = functools.partial(sh.map_list, ['calibrated_models', 'data'])


def select_predictions(outputs, targets, results):
    """
    Select the prediction results.

    :param outputs:
        Output keys.
    :type outputs: list | tuple

    :param targets:
        Targets keys.
    :type targets: list | tuple

    :param results:
        Calibration results.
    :type results: dict

    :return:
        Prediction results.
    :rtype: dict
    """
    results = sh.selector(outputs, results, allow_miss=True)
    return sh.map_dict(dict(zip(outputs, targets)), results)


def calculate_errors(references, predictions, metrics, metrics_kwargs):
    """
    Calculate error coefficients.

    :param references:
        Reference results.
    :type references: dict

    :param predictions:
        Prediction results.
    :type predictions: dict

    :param metrics:
        Metric functions.
    :type metrics: dict[str, function]

    :param metrics_kwargs:
        Kwargs to pass to metric functions.
    :type metrics_kwargs: dict

    :return:
        Error coefficients.
    :rtype: dict
    """
    it = set(predictions).intersection(references).intersection(metrics)
    try:
        return {
            k: metrics[k](references[k], predictions[k], **metrics_kwargs)
            for k in it
        }
    except TypeError:
        return {}


def _check_limit(limit, errors, check=lambda e, l: e <= l):
    it = set(limit).intersection(errors)
    return {k: check(errors[k], limit[k]) for k in it}


def calculate_calibration_status(errors, up_limit=None, dn_limit=None):
    """
    Calculate calibration statuses.

    :param errors:
        Error coefficients.
    :type errors: dict

    :param up_limit:
        Upper limits.
    :type up_limit: dict

    :param dn_limit:
        Lower limits.
    :type dn_limit: dict

    :return:
        Calibration statuses.
    :rtype: dict
    """
    status = {}

    limit = _check_limit(up_limit or {}, errors, check=lambda e, l: e <= l)
    if limit:
        status['up_limit'] = limit

    limit = _check_limit(dn_limit or {}, errors, check=lambda e, l: e >= l)
    if limit:
        status['up_limit'] = limit

    return status


def _mdl_error(mdl):
    dsp = sh.BlueDispatcher(
        name=mdl.name,
        description='Calculates the error of calibrated model of a reference.',
    )
    dsp.add_data('inputs_map', getattr(mdl, 'inputs_map', {}))

    dsp.add_function(
        function_id='select_inputs',
        function=sh.map_dict,
        inputs=['inputs_map', 'data'],
        outputs=['inputs<0>']
    )

    dsp.add_data('inputs', getattr(mdl, 'inputs', []))
    dsp.add_function(
        function_id='select_inputs',
        function=_select_data,
        inputs=['inputs', 'inputs<0>'],
        outputs=['inputs<1>']
    )

    dsp.add_function(
        function=sh.combine_dicts,
        inputs=['calibrated_models', 'inputs<1>'],
        outputs=['prediction_inputs']
    )

    dsp.add_data('targets', getattr(mdl, 'targets', []))
    dsp.add_function(
        function_id='select_targets', function=_select_data,
        inputs=['targets', 'data'], outputs=['references']
    )

    dsp.add_function(
        function=sh.SubDispatch(mdl.dsp),
        inputs=['prediction_inputs', 'calibrated_models'],
        outputs=['results']
    )

    dsp.add_data('outputs', getattr(mdl, 'outputs', []))
    dsp.add_func(select_predictions, outputs=['predictions'])

    dsp.add_data('metrics_inputs', getattr(mdl, 'metrics_inputs', {}))
    dsp.add_function(
        function_id='select_metrics_inputs',
        function=_select_data,
        inputs=['metrics_inputs', 'data'],
        outputs=['metrics_kwargs']
    )

    dsp.add_data('metrics', getattr(mdl, 'metrics', {}))
    dsp.add_func(calculate_errors, outputs=['errors'])

    dsp.add_data('up_limit', getattr(mdl, 'up_limit', None))
    dsp.add_data('dn_limit', getattr(mdl, 'dn_limit', None))
    dsp.add_func(
        calculate_calibration_status, inputs_kwargs=True, outputs=['status']
    )

    return dsp


def _mdl_errors(mdl, data_id, err_func):
    name = '%s-%s errors' % (mdl.name, data_id)
    dsp = sh.BlueDispatcher(
        name=name, description='Calculates the error of calibrated model.'
    )
    dsp.add_data('models', mdl.models)

    dsp.add_function(
        function_id='select_models',
        function=getattr(mdl, 'select_models', _select_data),
        inputs=['models', data_id],
        outputs=['calibrated_models']
    )
    dsp.add_data('data_in', data_id)

    for o in calibration_cycles:
        dsp.add_function(
            function=_map_list,
            inputs=['calibrated_models', o],
            outputs=['input/%s' % o]
        )

        dsp.add_function(
            function=err_func,
            inputs=['input/%s' % o],
            outputs=['error/%s' % o]
        )
    return dsp, name


def _mean(values, weights=None):
    if isinstance(weights, Iterable):
        values = [v * w for v, w in zip(values, weights) if w]

    v = np.asarray(values)
    return np.average(v)


def _key_score(x):
    s = 1 if np.isnan(x['score']) else -int(x['success'])
    return s, -x['n'], x['score']


def _key_scores(x):
    return tuple(y[:2] for y in x)


def _sorting_func(x):
    return _key_score(x[0]) + _key_scores(x[1]) + (x[3],)


def sort_models(*data, weights=None):
    """
    Sort models by scores.

    :param weights:
        Weights coefficients.
    :type weights: dict

    :return:
        Models rank.
    :rtype: list
    """
    weights = weights or {}
    rank = []

    for d in data:
        errors = {k[6:]: v for k, v in d.items() if k.startswith('error/')}
        scores = []

        def _sort(x):
            return x[0], x[1], tuple(x[2].values()), x[3]

        for k, v in errors.items():
            if v[0]:
                l = [list(m.values()) for l, m in sorted(v[1].items()) if m]
                l = _mean(l) if l else 1
                keys, m = zip(*v[0].items())
                e = l, _mean(m, weights=[weights.get(i, 1) for i in keys])
                scores.append((e, l, v[0], k, v[1]))

        scores = sorted(scores, key=_sort)
        if scores:
            score = tuple(np.mean([e[0] for e in scores], axis=0))

            models = d['calibrated_models']

            if models:
                score = {
                    'success': score[0] == 1,
                    'n': len(models),
                    'score': score[1]
                }

            rank.append([score, scores, errors, d['data_in'], models])

    return sorted(rank, key=_sorting_func)


def _check_success(score):
    try:
        return score['success']
    except IndexError:
        return True


def format_score(rank):
    """
    Format score output.

    :param rank:
        Models rank.
    :type rank: list

    :return:
        Score output.
    :rtype: dict
    """
    res = {}
    for score, scores, errors, name, models in rank:
        res[name] = d = {'models': sorted(models.keys())}
        if scores:
            d.update({
                'score': score,
                'errors': {k: v[0] for k, v in errors.items()},
                'limits': {k: v[1] for k, v in errors.items()}
            })
    return res


def select_best_model(rank, enable_selector, selector_id=''):
    """
    Select the best model.

    :param rank:
        Models rank.
    :type rank: list

    :param enable_selector:
        Enable the selection of the best model to predict both H/L cycles.
    :type enable_selector: bool

    :param selector_id:
        Selector id.
    :type selector_id: str

    :return:
        Best model.
    :rtype: dict
    """

    models, _map = {}, {}
    if not enable_selector:
        from co2mpas.defaults import dfl
        _map = dfl.functions.select_best_model.MAP
        _map = _map.get(selector_id, _map[None])

    for i, (score, scores, err, name, mdl) in enumerate(rank):
        k = ((sh.NONE,) if enable_selector and i == 0 else ())
        k += _map.get(name, ())
        if not k:
            continue
        success = _check_success(score)
        models.update(dict.fromkeys(k, (name, success, mdl)))
        if not success:
            log.info(
                '\n  %s: Models (%s) to predict %s failed the calibration.',
                '%s warning' % selector_id.replace('_', ' ').capitalize(),
                ','.join(sorted(set(mdl))), ','.join(k)
            )
    return models


def mdl_selector(name, package=None):
    """
    Defines a model selector for a specific model.

    :param name:
        Model name.
    :type name: str

    :param package:
        Package name.
    :type package: str

    :return:
        Model selector.
    :rtype: schedula.utils.blue.BlueDispatcher
    """
    import importlib
    mdl = importlib.import_module(name, package)
    dsp = sh.BlueDispatcher(
        name='%s selector' % mdl.name,
        description='Select the calibrated %s.' % mdl.name
    )

    err_func = sh.SubDispatch(
        _mdl_error(mdl), outputs=['errors', 'status'], output_type='list'
    )

    for data_id in calibration_cycles:
        mdl_err, mdl_err_name = _mdl_errors(mdl, data_id, err_func)
        dsp.add_func(
            sh.SubDispatchFunction(
                mdl_err, function_id=mdl_err_name,
                inputs=[data_id] + [
                    k for k in calibration_cycles if k != data_id
                ]
            ),
            outputs=['error/%s' % data_id]
        )

    # noinspection PyTypeChecker
    dsp.add_function(
        function=functools.partial(
            sort_models, weights=getattr(mdl, 'weights', None)
        ),
        inputs=list(map('error/{}'.format, calibration_cycles)),
        outputs=['rank']
    )

    dsp.add_func(format_score, outputs=['score'])

    dsp.add_func(functools.partial(
        select_best_model, selector_id='%s selector' % mdl.name
    ), outputs=['model'])

    return dsp



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the idle engine speed.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl

dsp = sh.BlueDispatcher(
    name='calculate_idle_engine_speed',
    description='Identify idle engine speed median and std.'
)
dsp.add_data('idle_engine_speed_std', dfl.values.idle_engine_speed_std, 20)


@sh.add_function(dsp, outputs=['idle_model_detector'], weight=100)
def define_idle_model_detector(
        velocities, engine_speeds_out, stop_velocity, min_engine_on_speed):
    """
    Defines idle engine speed model detector.

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :return:
        Idle engine speed model detector.
    :rtype: sklearn.cluster.DBSCAN
    """

    b = (velocities < stop_velocity) & (engine_speeds_out > min_engine_on_speed)
    if not b.any():
        return sh.NONE
    x = engine_speeds_out[b, None]
    from ._idle import _IdleDetector
    model = _IdleDetector(eps=dfl.functions.define_idle_model_detector.EPS)
    model.fit(x)
    return model


@sh.add_function(dsp, outputs=['idle_engine_speed_median'])
def identify_idle_engine_speed_median(idle_model_detector):
    """
    Identifies idle engine speed [RPM].

    :param idle_model_detector:
        Idle engine speed model detector.
    :type idle_model_detector: _IdleDetector

    :return:
        Idle engine speed [RPM].
    :rtype: float
    """
    imd = idle_model_detector
    return np.median(imd.cluster_centers_[imd.labels_])


@sh.add_function(dsp, outputs=['idle_engine_speed_std'])
def identify_idle_engine_speed_std(
        idle_model_detector, engine_speeds_out, idle_engine_speed_median,
        min_engine_on_speed):
    """
    Identifies standard deviation of idle engine speed [RPM].

    :param idle_model_detector:
        Idle engine speed model detector.
    :type idle_model_detector: _IdleDetector

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param idle_engine_speed_median:
        Idle engine speed [RPM].
    :type idle_engine_speed_median: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :return:
        Standard deviation of idle engine speed [RPM].
    :rtype: float
    """
    b = idle_model_detector.predict([(idle_engine_speed_median,)],
                                    set_outliers=False)
    b = idle_model_detector.predict(engine_speeds_out[:, None]) == b
    b &= (engine_speeds_out > min_engine_on_speed)
    idle_std = dfl.functions.identify_idle_engine_speed_std.MIN_STD
    # noinspection PyUnresolvedReferences
    if not b.any():
        return idle_std

    s = np.sqrt(np.mean((engine_speeds_out[b] - idle_engine_speed_median) ** 2))

    p = dfl.functions.identify_idle_engine_speed_std.MAX_STD_PERC
    return min(max(s, idle_std), idle_engine_speed_median * p)


dsp.add_function(
    function=sh.bypass,
    inputs=['idle_engine_speed_median', 'idle_engine_speed_std'],
    outputs=['idle_engine_speed']
)

dsp.add_function(
    function=sh.bypass,
    inputs=['idle_engine_speed'],
    outputs=['idle_engine_speed_median', 'idle_engine_speed_std']
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the electric motor in P3 position.
"""
import schedula as sh

dsp = sh.BlueDispatcher(
    name='Motor P3',
    description='Models the motor P3 (motor upstream the final drive).'
)


@sh.add_function(
    dsp, inputs=['motor_p3_front_powers'],
    outputs=['motor_p3_front_maximum_power']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_powers'],
    outputs=['motor_p3_rear_maximum_power']
)
def identify_motor_p3_maximum_power(motor_p3_powers):
    """
    Identify the maximum power of motor P3 [kW].

    :param motor_p3_powers:
        Power at motor P3 [kW].
    :type motor_p3_powers: numpy.array

    :return:
        Maximum power of motor P3 [kW].
    :rtype: float
    """
    from .p4 import identify_motor_p4_maximum_power as func
    return func(motor_p3_powers)


@sh.add_function(
    dsp, inputs=['motor_p3_front_torques'],
    outputs=['motor_p3_front_maximum_torque']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_torques'],
    outputs=['motor_p3_rear_maximum_torque']
)
def identify_motor_p3_maximum_torque(motor_p3_torques):
    """
    Identify the maximum torque of motor P3 [N*m].

    :param motor_p3_torques:
        Torque at motor P3 [N*m].
    :type motor_p3_torques: numpy.array

    :return:
        Maximum torque of motor P3 [N*m].
    :rtype: float
    """
    from .p4 import identify_motor_p4_maximum_torque as func
    return func(motor_p3_torques)


@sh.add_function(
    dsp, inputs=['motor_p3_front_rated_speed', 'motor_p3_front_maximum_torque'],
    outputs=['motor_p3_front_maximum_power']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_rated_speed', 'motor_p3_rear_maximum_torque'],
    outputs=['motor_p3_rear_maximum_power']
)
def calculate_motor_p3_maximum_power(
        motor_p3_rated_speed, motor_p3_maximum_torque):
    """
    Calculate the maximum power of motor P3 [kW].

    :param motor_p3_rated_speed:
        Rated speed of motor P3 [RPM].
    :type motor_p3_rated_speed: float

    :param motor_p3_maximum_torque:
        Maximum torque of motor P3 [N*m].
    :type motor_p3_maximum_torque: float

    :return:
        Maximum power of motor P3 [kW].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_maximum_power as func
    return func(motor_p3_rated_speed, motor_p3_maximum_torque)


@sh.add_function(
    dsp, outputs=['motor_p3_front_rated_speed'],
    inputs=['motor_p3_front_maximum_power', 'motor_p3_front_maximum_torque']
)
@sh.add_function(
    dsp, outputs=['motor_p3_rear_rated_speed'],
    inputs=['motor_p3_rear_maximum_power', 'motor_p3_rear_maximum_torque']
)
def calculate_motor_p3_rated_speed(
        motor_p3_maximum_power, motor_p3_maximum_torque):
    """
    Calculate the rated speed of motor P3 [RPM].

    :param motor_p3_maximum_power:
        Maximum power of motor P3 [kW].
    :type motor_p3_maximum_power: float

    :param motor_p3_maximum_torque:
        Maximum torque of motor P3 [N*m].
    :type motor_p3_maximum_torque: float

    :return:
        Rated speed of motor P3 [RPM].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_rated_speed as func
    return func(motor_p3_maximum_power, motor_p3_maximum_torque)


@sh.add_function(
    dsp, outputs=['motor_p3_front_maximum_torque'],
    inputs=['motor_p3_front_maximum_power', 'motor_p3_front_rated_speed']
)
@sh.add_function(
    dsp, outputs=['motor_p3_rear_maximum_torque'],
    inputs=['motor_p3_rear_maximum_power', 'motor_p3_rear_rated_speed']
)
def calculate_motor_p3_maximum_torque(
        motor_p3_maximum_power, motor_p3_rated_speed):
    """
    Calculate the maximum torque of motor P3 [N*m].

    :param motor_p3_maximum_power:
        Maximum power of motor P3 [kW].
    :type motor_p3_maximum_power: float

    :param motor_p3_rated_speed:
        Rated speed of motor P3 [RPM].
    :type motor_p3_rated_speed: float

    :return:
        Maximum torque of motor P3 [N*m].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_maximum_torque as func
    return func(motor_p3_maximum_power, motor_p3_rated_speed)


@sh.add_function(
    dsp, inputs=['motor_p3_front_maximum_power', 'motor_p3_front_rated_speed'],
    outputs=['motor_p3_front_maximum_power_function']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_maximum_power', 'motor_p3_rear_rated_speed'],
    outputs=['motor_p3_rear_maximum_power_function']
)
def define_motor_p3_maximum_power_function(
        motor_p3_maximum_power, motor_p3_rated_speed):
    """
    Define the maximum power function of motor P3.

    :param motor_p3_maximum_power:
        Maximum power of motor P3 [kW].
    :type motor_p3_maximum_power: float

    :param motor_p3_rated_speed:
        Rated speed of motor P3 [RPM].
    :type motor_p3_rated_speed: float

    :return:
        Maximum power function of motor P3.
    :rtype: function
    """
    from .p4 import define_motor_p4_maximum_power_function as func
    return func(motor_p3_maximum_power, motor_p3_rated_speed)


@sh.add_function(
    dsp, outputs=['motor_p3_front_maximum_powers'],
    inputs=['motor_p3_front_speeds', 'motor_p3_front_maximum_power_function']
)
@sh.add_function(
    dsp, outputs=['motor_p3_rear_maximum_powers'],
    inputs=['motor_p3_rear_speeds', 'motor_p3_rear_maximum_power_function']
)
def calculate_motor_p3_maximum_powers(
        motor_p3_speeds, motor_p3_maximum_power_function):
    """
    Calculate the maximum power vector of motor P3 [kW].

    :param motor_p3_speeds:
        Rotating speed of motor P3 [RPM].
    :type motor_p3_speeds: numpy.array | float

    :param motor_p3_maximum_power_function:
        Maximum power function of motor P3.
    :type motor_p3_maximum_power_function: function

    :return:
        Maximum power vector of motor P3 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_maximum_powers as func
    return func(motor_p3_speeds, motor_p3_maximum_power_function)


@sh.add_function(
    dsp, inputs=['final_drive_speeds_in', 'motor_p3_front_speeds'],
    outputs=['motor_p3_front_speed_ratio']
)
@sh.add_function(
    dsp, inputs=['final_drive_speeds_in', 'motor_p3_rear_speeds'],
    outputs=['motor_p3_rear_speed_ratio']
)
def identify_motor_p3_speed_ratio(final_drive_speeds_in, motor_p3_speeds):
    """
    Identifies motor P3 speed ratio.

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :param motor_p3_speeds:
        Rotating speed of motor P3 [RPM].
    :type motor_p3_speeds: numpy.array | float

    :return:
        Motor P3 speed ratio [-].
    :rtype: float
    """
    from .p4 import identify_motor_p4_speed_ratio as func
    return func(final_drive_speeds_in, motor_p3_speeds)


dsp.add_data('motor_p3_front_speed_ratio', 1, sh.inf(10, 1))
dsp.add_data('motor_p3_rear_speed_ratio', 1, sh.inf(10, 1))


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['motor_p3_front_speeds'],
    inputs=['final_drive_speeds_in', 'motor_p3_front_speed_ratio']
)
@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['motor_p3_rear_speeds'],
    inputs=['final_drive_speeds_in', 'motor_p3_rear_speed_ratio']
)
def calculate_motor_p3_speeds(final_drive_speeds_in, motor_p3_speed_ratio=1):
    """
    Calculates rotating speed of motor P3 [RPM].

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :param motor_p3_speed_ratio:
        Ratio between motor P3 speed and final drive speed in [-].
    :type motor_p3_speed_ratio: float

    :return:
        Rotating speed of motor P3 [RPM].
    :rtype: numpy.array | float
    """
    return final_drive_speeds_in * motor_p3_speed_ratio


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['final_drive_speeds_in'],
    inputs=['motor_p3_front_speeds', 'motor_p3_front_speed_ratio']
)
@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['final_drive_speeds_in'],
    inputs=['motor_p3_rear_speeds', 'motor_p3_rear_speed_ratio']
)
def calculate_final_drive_speeds_in(motor_p3_speeds, motor_p3_speed_ratio=1):
    """
    Calculates final drive speed [RPM].

    :param motor_p3_speeds:
        Rotating speed of motor P3 [RPM].
    :type motor_p3_speeds: numpy.array | float

    :param motor_p3_speed_ratio:
        Ratio between motor P3 speed and final drive speed in [-].
    :type motor_p3_speed_ratio: float

    :return:
        Final drive speed in [RPM].
    :rtype: numpy.array | float
    """
    return motor_p3_speeds / motor_p3_speed_ratio


@sh.add_function(
    dsp, inputs=['motor_p3_front_powers', 'motor_p3_front_speeds'],
    outputs=['motor_p3_front_torques']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_powers', 'motor_p3_rear_speeds'],
    outputs=['motor_p3_rear_torques']
)
def calculate_motor_p3_torques(motor_p3_powers, motor_p3_speeds):
    """
    Calculates torque at motor P3 [N*m].

    :param motor_p3_powers:
        Power at motor P3 [kW].
    :type motor_p3_powers: numpy.array | float

    :param motor_p3_speeds:
        Rotating speed of motor P3 [RPM].
    :type motor_p3_speeds: numpy.array | float

    :return:
        Torque at motor P3 [N*m].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques
    return calculate_wheel_torques(motor_p3_powers, motor_p3_speeds)


@sh.add_function(
    dsp, inputs=['motor_p3_front_torques', 'motor_p3_front_speeds'],
    outputs=['motor_p3_front_powers']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_torques', 'motor_p3_rear_speeds'],
    outputs=['motor_p3_rear_powers']
)
def calculate_motor_p3_powers(motor_p3_torques, motor_p3_speeds):
    """
    Calculates power at motor P3 [kW].

    :param motor_p3_torques:
        Torque at motor P3 [N*m].
    :type motor_p3_torques: numpy.array | float

    :param motor_p3_speeds:
        Rotating speed of motor P3 [RPM].
    :type motor_p3_speeds: numpy.array | float

    :return:
        Power at motor P3 [kW].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_powers
    return calculate_wheel_powers(motor_p3_torques, motor_p3_speeds)


@sh.add_function(
    dsp, inputs=['motor_p3_front_powers', 'motor_p3_front_torques'],
    outputs=['motor_p3_front_speeds']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_powers', 'motor_p3_rear_torques'],
    outputs=['motor_p3_rear_speeds']
)
def calculate_motor_p3_speeds_v1(motor_p3_powers, motor_p3_torques):
    """
    Calculates rotating speed of motor P3 [RPM].

    :param motor_p3_powers:
        Power at motor P3 [kW].
    :type motor_p3_powers: numpy.array | float

    :param motor_p3_torques:
        Torque at motor P3 [N*m].
    :type motor_p3_torques: numpy.array | float

    :return:
        Rotating speed of motor P3 [RPM].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques
    return calculate_wheel_torques(motor_p3_powers, motor_p3_torques)


dsp.add_data('motor_p3_front_efficiency', 0.9)
dsp.add_data('motor_p3_rear_efficiency', 0.9)


@sh.add_function(
    dsp, inputs=['motor_p3_front_powers', 'motor_p3_front_efficiency'],
    outputs=['motor_p3_front_electric_powers']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_powers', 'motor_p3_rear_efficiency'],
    outputs=['motor_p3_rear_electric_powers']
)
def calculate_motor_p3_electric_powers(motor_p3_powers, motor_p3_efficiency):
    """
    Calculates motor P3 electric power [kW].

    :param motor_p3_powers:
        Power at motor P3 [kW].
    :type motor_p3_powers: numpy.array | float

    :param motor_p3_efficiency:
        Motor P3 efficiency [-].
    :type motor_p3_efficiency: float

    :return:
        Electric power of motor P3 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_electric_powers as func
    return func(motor_p3_powers, motor_p3_efficiency)


@sh.add_function(
    dsp, inputs=['motor_p3_front_electric_powers', 'motor_p3_front_efficiency'],
    outputs=['motor_p3_front_powers']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_electric_powers', 'motor_p3_rear_efficiency'],
    outputs=['motor_p3_rear_powers']
)
def calculate_motor_p3_powers_v1(motor_p3_electric_powers, motor_p3_efficiency):
    """
    Calculate motor P3 power from electric power and electric power losses [kW].

    :param motor_p3_electric_powers:
        Electric power of motor P3 [kW].
    :type motor_p3_electric_powers: numpy.array | float

    :param motor_p3_efficiency:
        Motor P3 efficiency [-].
    :type motor_p3_efficiency: float

    :return:
        Power at motor P3 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_powers_v1 as func
    return func(motor_p3_electric_powers, motor_p3_efficiency)


dsp.add_data('has_motor_p3_front', False, sh.inf(10, 3))
dsp.add_data('has_motor_p3_rear', False, sh.inf(10, 3))


@sh.add_function(
    dsp, inputs=['motor_p3_front_maximum_power'], outputs=['has_motor_p3_front']
)
@sh.add_function(
    dsp, inputs=['motor_p3_rear_maximum_power'], outputs=['has_motor_p3_rear']
)
def identify_has_motor_p3(motor_p3_maximum_power):
    """
    Identify if the vehicle has a motor P3 [kW].

    :param motor_p3_maximum_power:
        Maximum power of motor P3 [kW].
    :type motor_p3_maximum_power: float

    :return:
        Has the vehicle a motor in P3?
    :rtype: bool
    """
    from .p4 import identify_has_motor_p4 as func
    return func(motor_p3_maximum_power)


@sh.add_function(
    dsp, inputs=['times', 'has_motor_p3_front'],
    outputs=['motor_p3_front_powers']
)
@sh.add_function(
    dsp, inputs=['times', 'has_motor_p3_rear'], outputs=['motor_p3_rear_powers']
)
def default_motor_p3_powers(times, has_motor_p3):
    """
    Return zero power if the vehicle has not a motor P3 [kW].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param has_motor_p3:
        Has the vehicle a motor in P3?
    :type has_motor_p3: bool

    :return:
        Power at motor P3 [kW].
    :rtype: numpy.array
    """
    from .p4 import default_motor_p4_powers as func
    return func(times, has_motor_p3)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the legislation corrections.
"""
import functools
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='Legislation', description='Models the legislation corrections.'
)


@sh.add_function(dsp, outputs=['phases_indices'])
def identify_phases_indices(times, phases_integration_times):
    """
    Identifies the indices of the cycle phases [-].

    If phases_distances is not given the result is the cumulative CO2 of cycle
    phases [CO2g] otherwise it is CO2 emission of cycle phases [CO2g/km].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param phases_integration_times:
        Cycle phases integration times [s].
    :type phases_integration_times: tuple

    :return:
        Indices of the cycle phases [-].
    :rtype: numpy.array
    """
    return np.searchsorted(times, phases_integration_times, side='right')


@sh.add_function(dsp, inputs_kwargs=True, outputs=['phases_co2_emissions'])
def calculate_phases_co2_emissions(
        times, phases_indices, co2_emissions, phases_distances=1.0):
    """
    Calculates CO2 emission or cumulative CO2 of cycle phases [CO2g/km or CO2g].

    If phases_distances is not given the result is the cumulative CO2 of cycle
    phases [CO2g] otherwise it is CO2 emission of cycle phases [CO2g/km].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :param co2_emissions:
        CO2 instantaneous emissions vector [CO2g/s].
    :type co2_emissions: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array | float, optional

    :return:
        CO2 emission or cumulative CO2 of cycle phases [CO2g/km or CO2g].
    :rtype: numpy.array
    """
    from scipy.integrate import trapz
    co2 = []
    for i, j in phases_indices:
        co2.append(trapz(co2_emissions[i:j], times[i:j]))

    with np.errstate(divide='ignore', invalid='ignore'):
        return np.nan_to_num(np.array(co2) / phases_distances)


@sh.add_function(dsp, outputs=['co2_emission_value'])
def calculate_co2_emission_value(phases_co2_emissions, phases_distances):
    """
    Calculates the CO2 emission of the cycle [CO2g/km].

    :param phases_co2_emissions:
        CO2 emission of cycle phases [CO2g/km].
    :type phases_co2_emissions: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array | float

    :return:
        CO2 emission value of the cycle [CO2g/km].
    :rtype: float
    """

    n = sum(phases_co2_emissions * phases_distances)

    if isinstance(phases_distances, float):
        d = phases_distances * len(phases_co2_emissions)
    else:
        d = sum(phases_distances)

    return float(n / d)


dsp.add_data(
    'has_periodically_regenerating_systems',
    dfl.values.has_periodically_regenerating_systems
)
dsp.add_data('ki_additive', dfl.values.ki_additive)


@sh.add_function(dsp, outputs=['ki_multiplicative'])
def default_ki_multiplicative(
        has_periodically_regenerating_systems, ki_additive):
    """
    Returns the default ki multiplicative factor [-].

    :param has_periodically_regenerating_systems:
        Does the vehicle has periodically regenerating systems? [-].
    :type has_periodically_regenerating_systems: bool

    :param ki_additive:
        Additive correction for vehicles with periodically regenerating
        systems [CO2g/km].
    :type ki_additive: float

    :return:
        Multiplicative correction for vehicles with periodically regenerating
        systems [-].
    :rtype: float
    """
    if ki_additive:
        return 1.0
    par = dfl.functions.default_ki_multiplicative.ki_multiplicative
    return par.get(has_periodically_regenerating_systems, 1.0)


@sh.add_function(dsp, outputs=['phases_co2_emissions'])
def merge_wltp_phases_co2_emission(
        co2_emission_low, co2_emission_medium, co2_emission_high,
        co2_emission_extra_high):
    """
    Merges WLTP phases co2 emission.

    :param co2_emission_low:
        CO2 emission on low WLTP phase [CO2g/km].
    :type co2_emission_low: float

    :param co2_emission_medium:
        CO2 emission on medium WLTP phase [CO2g/km].
    :type co2_emission_medium: float

    :param co2_emission_high:
        CO2 emission on high WLTP phase [CO2g/km].
    :type co2_emission_high: float

    :param co2_emission_extra_high:
        CO2 emission on extra high WLTP phase [CO2g/km].
    :type co2_emission_extra_high: float

    :return:
        CO2 emission of cycle phases [CO2g/km].
    :rtype: tuple[float]
    """
    return (co2_emission_low, co2_emission_medium, co2_emission_high,
            co2_emission_extra_high)


dsp.add_function(
    function_id='split_wltp_phases_co2_emission',
    function=sh.add_args(sh.bypass),
    inputs=['cycle_type', 'phases_co2_emissions'],
    outputs=[
        'co2_emission_low', 'co2_emission_medium', 'co2_emission_high',
        'co2_emission_extra_high'
    ],
    input_domain=lambda cycle, phases: (cycle, len(phases)) == ('WLTP', 4)
)

dsp.add_function(
    function_id='split_nedc_phases_co2_emission',
    function=sh.add_args(sh.bypass),
    inputs=['cycle_type', 'phases_co2_emissions'],
    outputs=['co2_emission_UDC', 'co2_emission_EUDC'],
    input_domain=lambda cycle, phases: (cycle, len(phases)) == ('NEDC', 2)
)


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['phases_co2_emissions'])
def merge_nedc_phases_co2_emission(co2_emission_UDC, co2_emission_EUDC):
    """
    Merges NEDC phases co2 emission.

    :param co2_emission_UDC:
        CO2 emission on UDC NEDC phase [CO2g/km].
    :type co2_emission_UDC: float

    :param co2_emission_EUDC:
        CO2 emission on EUDC NEDC phase [CO2g/km].
    :type co2_emission_EUDC: float

    :return:
        CO2 emission of cycle phases [CO2g/km].
    :rtype: tuple[float]
    """
    return co2_emission_UDC, co2_emission_EUDC


def _rcb_correction(
        phases_delta_energy, phases_distances, fuel_type, engine_type,
        alternator_efficiency):
    # noinspection PyProtectedMember
    par = dfl.functions._rcb_correction.WILLANS
    dco2 = np.nan_to_num(-phases_delta_energy / phases_distances)
    dco2 *= par[fuel_type][engine_type] * 0.0036 / alternator_efficiency
    return dco2


@sh.add_function(dsp, outputs=['theoretical_phases_distances'])
def calculate_theoretical_phases_distances(
        times, theoretical_velocities, phases_indices):
    """
    Calculates theoretical cycle phases distances [km].

    :param times:
        Time vector.
    :type times: numpy.array

    :param theoretical_velocities:
        Theoretical velocity vector [km/h].
    :type theoretical_velocities: numpy.array

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :return:
        Theoretical cycle phases distances [km].
    :rtype:  numpy.array
    """
    from .vehicle import calculate_distances
    return calculate_phases_distances(
        phases_indices, calculate_distances(times, theoretical_velocities)
    )


dsp.add_data('rcb_correction', dfl.values.rcb_correction)


@sh.add_function(dsp, outputs=['speed_distance_correction'])
def default_speed_distance_correction(is_hybrid):
    """
    Returns if speed distance correction has to be applied.

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        Apply speed distance correction?
    :rtype: bool
    """
    return not is_hybrid


@sh.add_function(
    dsp, inputs_kwargs=True,
    outputs=['speed_distance_corrected_co2_emission_value']
)
def calculate_speed_distance_corrected_co2_emission_value(
        phases_co2_emissions, phases_times, batteries_phases_delta_energy,
        theoretical_phases_distances, phases_distances, alternator_efficiency,
        engine_type, fuel_type, motive_powers, theoretical_motive_powers, times,
        phases_indices, engine_max_power, speed_distance_correction=True,
        is_hybrid=False, cycle_type='WLTP'):
    """
    Calculates the CO2 emission value corrected for speed & distance [CO2g/km].

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param phases_co2_emissions:
        CO2 emission of cycle phases [CO2g/km].
    :type phases_co2_emissions: numpy.array

    :param phases_times:
        Cycle phases times [s].
    :type phases_times: numpy.array

    :param batteries_phases_delta_energy:
        Phases delta energy of the batteries [Wh].
    :type batteries_phases_delta_energy: numpy.array

    :param theoretical_phases_distances:
        Theoretical cycle phases distances [km].
    :type theoretical_phases_distances: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array

    :param alternator_efficiency:
        Alternator efficiency [-].
    :type alternator_efficiency: float

    :param engine_type:
        Engine type (positive turbo, positive natural aspiration, compression).
    :type engine_type: str

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param theoretical_motive_powers:
        Theoretical motive power [kW].
    :type theoretical_motive_powers: numpy.array

    :param times:
        Time vector.
    :type times: numpy.array

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :param speed_distance_correction:
        Apply speed distance correction?
    :type speed_distance_correction: bool

    :return:
        CO2 emission value corrected for speed & distance [CO2g/km].
    :rtype: float
    """
    if is_hybrid:
        return sh.NONE
    p_co2, p_dist = np.array(phases_co2_emissions), phases_distances
    if cycle_type == 'WLTP' and speed_distance_correction:
        p_co2 = np.column_stack((p_co2, _rcb_correction(
            batteries_phases_delta_energy, phases_distances, fuel_type,
            engine_type, alternator_efficiency
        ))) * np.nan_to_num(phases_distances / phases_times)[:, None]

        from sklearn.linear_model import LinearRegression
        mdl = LinearRegression()
        cpmp = functools.partial(
            calculate_phases_co2_emissions, times, phases_indices,
            phases_distances=phases_times
        )
        y, p_co2 = p_co2.sum(axis=1).ravel(), p_co2[:, 0].ravel()
        x = cpmp(np.maximum(-.02 * engine_max_power, motive_powers))
        mdl.fit(x[:, None], y)
        p2 = -mdl.intercept_ / mdl.coef_ if mdl.coef_ else -float('inf')
        dp_mp = cpmp(np.maximum(p2, motive_powers))
        mdl.fit(dp_mp[:, None], y)
        dp_mp -= cpmp(np.maximum(p2, theoretical_motive_powers))
        p_co2 -= mdl.coef_ * dp_mp
        p_co2 *= np.nan_to_num(phases_times / theoretical_phases_distances)
        p_dist = theoretical_phases_distances
    return calculate_co2_emission_value(p_co2, p_dist)


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['rcb_corrected_co2_emission_value']
)
def calculate_rcb_corrected_co2_emission_value(
        speed_distance_corrected_co2_emission_value, engine_type, fuel_type,
        alternator_efficiency, phases_distances, theoretical_phases_distances,
        batteries_phases_delta_energy, speed_distance_correction=True,
        rcb_correction=True, is_hybrid=False, cycle_type='WLTP'):
    """
    Calculates the CO2 emission value corrected for RCB [CO2g/km].

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param speed_distance_corrected_co2_emission_value:
        CO2 emission value corrected for speed & distance [CO2g/km].
    :type speed_distance_corrected_co2_emission_value: float

    :param batteries_phases_delta_energy:
        Phases delta energy of the batteries [Wh].
    :type batteries_phases_delta_energy: numpy.array

    :param theoretical_phases_distances:
        Theoretical cycle phases distances [km].
    :type theoretical_phases_distances: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array

    :param alternator_efficiency:
        Alternator efficiency [-].
    :type alternator_efficiency: float

    :param engine_type:
        Engine type (positive turbo, positive natural aspiration, compression).
    :type engine_type: str

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :param rcb_correction:
        Apply RCB correction?
    :type rcb_correction: bool

    :param speed_distance_correction:
        Apply speed distance correction?
    :type speed_distance_correction: bool

    :return:
        CO2 emission value corrected for RCB [CO2g/km].
    :rtype: float
    """
    if is_hybrid:
        return sh.NONE
    if cycle_type == 'WLTP' and rcb_correction:
        d = phases_distances
        if speed_distance_correction:
            d = theoretical_phases_distances
        return speed_distance_corrected_co2_emission_value + _rcb_correction(
            np.sum(batteries_phases_delta_energy), np.sum(d), fuel_type,
            engine_type, alternator_efficiency
        )
    return speed_distance_corrected_co2_emission_value


@sh.add_function(dsp, outputs=['batteries_phases_delta_energy'])
def calculate_batteries_phases_delta_energy(
        times, phases_indices, drive_battery_electric_powers,
        service_battery_electric_powers):
    """
    Calculates the phases delta energy of the batteries [Wh].

    :param times:
        Time vector.
    :type times: numpy.array

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param service_battery_electric_powers:
        Service battery electric power [kW].
    :type service_battery_electric_powers: numpy.array

    :return:
        Phases delta energy of the batteries [Wh].
    :rtype: numpy.array
    """
    from scipy.integrate import cumtrapz
    p = service_battery_electric_powers + drive_battery_electric_powers
    e = cumtrapz(p, times, initial=0)
    i = phases_indices.copy()
    i[:, 1] -= 1
    return np.diff(e[i], axis=1).ravel() / 3.6


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['kco2_wltp_correction_factor']
)
def identify_kco2_wltp_correction_factor(
        drive_battery_electric_powers, service_battery_electric_powers,
        co2_emissions, times, force_on_engine, after_treatment_warm_up_phases,
        velocities, is_hybrid=True):
    """
    Identifies the kco2 correction factor [g/Wh].

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param service_battery_electric_powers:
        Service battery electric power [kW].
    :type service_battery_electric_powers: numpy.array

    :param force_on_engine:
        Phases when engine is on because parallel mode is forced [-].
    :type force_on_engine: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param co2_emissions:
        CO2 instantaneous emissions vector [CO2g/s].
    :type co2_emissions: numpy.array

    :param times:
        Time vector.
    :type times: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        kco2 correction factor [g/Wh].
    :rtype: float
    """
    if not is_hybrid:
        return sh.NONE
    from scipy.integrate import cumtrapz
    from sklearn.linear_model import RANSACRegressor, Lasso
    b = ~(force_on_engine | after_treatment_warm_up_phases)
    e = np.where(
        b, drive_battery_electric_powers + service_battery_electric_powers, 0
    )
    e = cumtrapz(e, times, initial=0) / 3.6
    co2 = cumtrapz(np.where(b, co2_emissions, 0), times, initial=0)
    km = cumtrapz(np.where(b, velocities / 3.6, 0), times, initial=0) / 1000
    # noinspection PyTypeChecker
    it = co2_utl.sliding_window(list(zip(km, zip(km, e, co2))), 5)
    d = np.diff(np.array([(v[0][1], v[-1][1]) for v in it]), axis=1)[:, 0, :].T
    e, co2 = d[1:] / d[0]
    d0 = t0 = -float('inf')
    b, dkm = [], .5
    dt = dkm / np.mean(velocities) * 3600
    for i, (d, t) in enumerate(zip(km, times)):
        if d > d0 and t > t0:
            d0, t0 = d + dkm, t + dt
            b.append(i)
    b = np.array(b)
    m = RANSACRegressor(
        random_state=0,
        base_estimator=Lasso(random_state=0, positive=True)
    ).fit(e[b, None], co2[b])
    return float(m.estimator_.coef_)


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['rcb_corrected_co2_emission_value']
)
def calculate_rcb_corrected_co2_emission_value_v1(
        co2_emission_value, batteries_phases_delta_energy,
        kco2_wltp_correction_factor, phases_distances, rcb_correction=True,
        is_hybrid=True, cycle_type='WLTP'):
    """
    Calculates the CO2 emission value corrected for RCB [CO2g/km].

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param co2_emission_value:
        CO2 emission value of the cycle [CO2g/km].
    :type co2_emission_value: float

    :param batteries_phases_delta_energy:
        Phases delta energy of the batteries [Wh].
    :type batteries_phases_delta_energy: numpy.array

    :param kco2_wltp_correction_factor:
        kco2 WLTP correction factor [CO2g/Wh].
    :type kco2_wltp_correction_factor: float

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array

    :param rcb_correction:
        Apply RCB correction?
    :type rcb_correction: bool

    :return:
        CO2 emission value corrected for RCB [CO2g/km].
    :rtype: float
    """
    if not is_hybrid or cycle_type == 'NEDC':
        return sh.NONE

    if rcb_correction and kco2_wltp_correction_factor:
        de = np.sum(batteries_phases_delta_energy) / np.sum(phases_distances)
        return co2_emission_value - kco2_wltp_correction_factor * de
    return co2_emission_value


@sh.add_function(dsp, outputs=['kco2_nedc_correction_factor'])
def default_kco2_nedc_correction_factor(
        kco2_wltp_correction_factor, drive_battery_nominal_voltage, distances):
    """
    Returns the kco2 NEDC correction factor [CO2g/km/Ah].

    :param kco2_wltp_correction_factor:
        kco2 WLTP correction factor [CO2g/Wh].
    :type kco2_wltp_correction_factor: float

    :param drive_battery_nominal_voltage:
        Drive battery nominal voltage [V].
    :type drive_battery_nominal_voltage: float

    :param distances:
        Cumulative distance vector [m].
    :type distances: numpy.array

    :return:
        kco2 NEDC correction factor [CO2g/km/Ah].
    :rtype: float
    """
    d = distances[-1] / 1000
    return kco2_wltp_correction_factor * drive_battery_nominal_voltage / d


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['rcb_corrected_co2_emission_value']
)
def calculate_rcb_corrected_co2_emission_value_v2(
        co2_emission_value, drive_battery_delta_state_of_charge,
        drive_battery_capacity, kco2_nedc_correction_factor,
        rcb_correction=True, is_hybrid=True, cycle_type='NEDC'):
    """
    Calculates the CO2 emission value corrected for RCB [CO2g/km].

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param co2_emission_value:
        CO2 emission value of the cycle [CO2g/km].
    :type co2_emission_value: float

    :param drive_battery_delta_state_of_charge:
        Overall delta state of charge of the drive battery [%].
    :type drive_battery_delta_state_of_charge: numpy.array

    :param kco2_nedc_correction_factor:
        kco2 NEDC correction factor [CO2g/km/Ah].
    :type kco2_nedc_correction_factor: float

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :param rcb_correction:
        Apply RCB correction?
    :type rcb_correction: bool

    :return:
        CO2 emission value corrected for RCB [CO2g/km].
    :rtype: float
    """
    if not is_hybrid or cycle_type != 'NEDC':
        return sh.NONE

    if rcb_correction and kco2_nedc_correction_factor:
        di = drive_battery_delta_state_of_charge * drive_battery_capacity / 100
        return co2_emission_value - kco2_nedc_correction_factor * di
    return co2_emission_value


dsp.add_data(
    'atct_family_correction_factor', dfl.values.atct_family_correction_factor
)


def calculate_corrected_co2_emission(
        rcb_corrected_co2_emission_value, ki_multiplicative, ki_additive,
        atct_family_correction_factor=1.0):
    """
    Calculates the corrected CO2 emission of the cycle [CO2g/km].

    :param rcb_corrected_co2_emission_value:
        CO2 emission value corrected for RCB [CO2g/km].
    :type rcb_corrected_co2_emission_value: float

    :param ki_multiplicative:
        Multiplicative correction for vehicles with periodically regenerating
        systems [-].
    :type ki_multiplicative: float

    :param ki_additive:
        Additive correction for vehicles with periodically regenerating
        systems [CO2g/km].
    :type ki_multiplicative: float

    :param atct_family_correction_factor:
        Family correction factor for representative regional temperatures [-].
    :type atct_family_correction_factor: float

    :return:
        Corrected CO2 emission value of the cycle [CO2g/km].
    :rtype: float
    """
    v = rcb_corrected_co2_emission_value * ki_multiplicative + ki_additive
    return v * atct_family_correction_factor


@sh.add_function(dsp, True, True, outputs=['is_plugin'])
def default_is_plugin(input_type=None):
    """
    Returns if the vehicle is a plugin.

    :param input_type:
        Input file type.
    :type input_type: str

    :return:
        Is the vehicle a plugin?
    :rtype: bool
    """

    return str(input_type).upper() == 'OVC-HEV' or dfl.values.is_plugin


dsp.add_function(
    function=sh.add_args(calculate_corrected_co2_emission),
    inputs=[
        'is_plugin', 'rcb_corrected_co2_emission_value', 'ki_multiplicative',
        'ki_additive', 'atct_family_correction_factor'
    ],
    outputs=['corrected_co2_emission_value'],
    input_domain=co2_utl.check_first_arg_false
)
dsp.add_function(
    function=sh.add_args(calculate_corrected_co2_emission),
    inputs=[
        'is_plugin', 'rcb_corrected_co2_emission_value', 'ki_multiplicative',
        'ki_additive', 'atct_family_correction_factor'
    ],
    outputs=['corrected_sustaining_co2_emission_value'],
    input_domain=co2_utl.check_first_arg
)


# noinspection PyUnusedLocal
def _domain_calculate_corrected_co2_emission_for_conventional_nedc(
        cycle_type, is_hybrid, *args):
    return not is_hybrid and cycle_type == 'NEDC'


dsp.add_function(
    function_id='calculate_corrected_co2_emission_for_conventional_nedc',
    function=sh.add_args(calculate_corrected_co2_emission, n=2),
    inputs=['cycle_type', 'is_hybrid', 'co2_emission_value',
            'ki_multiplicative', 'ki_additive',
            'atct_family_correction_factor'],
    outputs=['corrected_co2_emission_value'],
    input_domain=_domain_calculate_corrected_co2_emission_for_conventional_nedc
)
dsp.add_function(
    function=sh.bypass,
    inputs=['corrected_co2_emission_value'],
    outputs=['declared_co2_emission_value']
)
dsp.add_function(
    function=sh.bypass,
    inputs=['corrected_sustaining_co2_emission_value'],
    outputs=['declared_sustaining_co2_emission_value']
)


def calculate_willans_factors(
        co2_params_calibrated, engine_fuel_lower_heating_value, engine_stroke,
        engine_capacity, min_engine_on_speed, fmep_model, engine_speeds_out,
        engine_powers_out, times, velocities, accelerations, motive_powers,
        engine_temperatures, missing_powers, angle_slopes):
    """
    Calculates the Willans factors.

    :param co2_params_calibrated:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type co2_params_calibrated: lmfit.Parameters

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param missing_powers:
        Missing engine power [kW].
    :type missing_powers: numpy.array

    :param angle_slopes:
        Angle slope vector [rad].
    :type angle_slopes: numpy.array

    :return:
        Willans factors:

        - av_velocities                         [km/h]
        - av_slope                              [rad]
        - distance                              [km]
        - init_temp                             [°C]
        - av_temp                               [°C]
        - end_temp                              [°C]
        - av_vel_pos_mov_pow                    [kw/h]
        - av_pos_motive_powers                  [kW]
        - sec_pos_mov_pow                       [s]
        - av_neg_motive_powers                  [kW]
        - sec_neg_mov_pow                       [s]
        - av_pos_accelerations                  [m/s2]
        - av_engine_speeds_out_pos_pow          [RPM]
        - av_pos_engine_powers_out              [kW]
        - engine_bmep_pos_pow                   [bar]
        - mean_piston_speed_pos_pow             [m/s]
        - fuel_mep_pos_pow                      [bar]
        - fuel_consumption_pos_pow              [g/sec]
        - willans_a                             [g/kWh]
        - willans_b                             [g/h]
        - specific_fuel_consumption             [g/kWh]
        - indicated_efficiency                  [-]
        - willans_efficiency                    [-]

    :rtype: dict
    """

    from .engine import calculate_mean_piston_speeds
    from .engine.fc import calculate_brake_mean_effective_pressures
    av = np.average

    w = np.zeros_like(times, dtype=float)
    t = (times[:-1] + times[1:]) / 2
    # noinspection PyUnresolvedReferences
    w[0], w[1:-1], w[-1] = t[0] - times[0], np.diff(t), times[-1] - t[-1]

    f = {
        'av_velocities': av(velocities, weights=w),  # [km/h]
        'av_slope': av(angle_slopes, weights=w),
        'has_sufficient_power': 1 - av(missing_powers != 0, weights=w),
        'max_power_required': max(engine_powers_out + missing_powers)
    }

    f['distance'] = f['av_velocities'] * (times[-1] - times[0]) / 3600.0  # [km]

    b = engine_powers_out >= 0
    if b.any():
        p = co2_params_calibrated.valuesdict()
        _w = w[b]
        av_s = av(engine_speeds_out[b], weights=_w)
        av_p = av(engine_powers_out[b], weights=_w)
        av_mp = av(missing_powers[b], weights=_w)

        n_p = calculate_brake_mean_effective_pressures(
            av_s, av_p, engine_capacity, min_engine_on_speed
        )
        n_s = calculate_mean_piston_speeds(av_s, engine_stroke)

        f_mep, wfa = fmep_model(p, n_s, n_p, 1)[:2]

        c = engine_capacity / engine_fuel_lower_heating_value * av_s
        fc = f_mep * c / 1200.0
        ieff = av_p / (fc * engine_fuel_lower_heating_value) * 1000.0

        willans_a = 3600000.0 / engine_fuel_lower_heating_value / wfa
        willans_b = fmep_model(p, n_s, 0, 1)[0] * c * 3.0

        sfc = willans_a + willans_b / av_p

        willans_eff = 3600000.0 / (sfc * engine_fuel_lower_heating_value)

        f.update({
            'av_engine_speeds_out_pos_pow': av_s,  # [RPM]
            'av_pos_engine_powers_out': av_p,  # [kW]
            'av_missing_powers_pos_pow': av_mp,  # [kW]
            'engine_bmep_pos_pow': n_p,  # [bar]
            'mean_piston_speed_pos_pow': n_s,  # [m/s]
            'fuel_mep_pos_pow': f_mep,  # [bar]
            'fuel_consumption_pos_pow': fc,  # [g/sec]
            'willans_a': willans_a,  # [g/kW]
            'willans_b': willans_b,  # [g]
            'specific_fuel_consumption': sfc,  # [g/kWh]
            'indicated_efficiency': ieff,  # [-]
            'willans_efficiency': willans_eff  # [-]
        })

    b = motive_powers > 0
    if b.any():
        _w = w[b]
        f['av_vel_pos_mov_pow'] = av(velocities[b], weights=_w)  # [km/h]
        f['av_pos_motive_powers'] = av(motive_powers[b], weights=_w)  # [kW]
        f['sec_pos_mov_pow'] = np.sum(_w)  # [s]

    b = accelerations > 0
    if b.any():
        _w = w[b]
        f['av_pos_accelerations'] = av(accelerations[b], weights=_w)  # [m/s2]

    b = motive_powers < 0
    if b.any():
        _w = w[b]
        f['av_neg_motive_powers'] = av(motive_powers[b], weights=_w)  # [kW]
        f['sec_neg_mov_pow'] = np.sum(_w)  # [s]

    f['init_temp'] = engine_temperatures[0]  # [°C]
    f['av_temp'] = av(engine_temperatures, weights=w)  # [°C]
    f['end_temp'] = engine_temperatures[-1]  # [°C]

    return f


dsp.add_data(
    'enable_willans', dfl.values.enable_willans,
    description='Enable the calculation of Willans coefficients for '
                'the cycle?'
)

dsp.add_function(
    function=sh.add_args(calculate_willans_factors),
    inputs=[
        'enable_willans', 'co2_params_calibrated',
        'engine_fuel_lower_heating_value', 'engine_stroke', 'engine_capacity',
        'min_engine_on_speed', 'fmep_model', 'engine_speeds_out',
        'engine_powers_out', 'times', 'velocities', 'accelerations',
        'motive_powers', 'engine_temperatures', 'missing_powers',
        'angle_slopes'
    ],
    outputs=['willans_factors'],
    input_domain=co2_utl.check_first_arg
)


def calculate_phases_willans_factors(
        params, engine_fuel_lower_heating_value, engine_stroke, engine_capacity,
        min_engine_on_speed, fmep_model, times, phases_indices,
        engine_speeds_out, engine_powers_out, velocities, accelerations,
        motive_powers, engine_temperatures, missing_powers,
        angle_slopes):
    """
    Calculates the Willans factors for each phase.

    :param params:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type params: lmfit.Parameters

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param missing_powers:
        Missing engine power [kW].
    :type missing_powers: numpy.array

    :param angle_slopes:
        Angle slope vector [rad].
    :type angle_slopes: numpy.array

    :return:
        Willans factors:

        - av_velocities                         [km/h]
        - av_slope                              [rad]
        - distance                              [km]
        - init_temp                             [°C]
        - av_temp                               [°C]
        - end_temp                              [°C]
        - av_vel_pos_mov_pow                    [kw/h]
        - av_pos_motive_powers                  [kW]
        - sec_pos_mov_pow                       [s]
        - av_neg_motive_powers                  [kW]
        - sec_neg_mov_pow                       [s]
        - av_pos_accelerations                  [m/s2]
        - av_engine_speeds_out_pos_pow          [RPM]
        - av_pos_engine_powers_out              [kW]
        - engine_bmep_pos_pow                   [bar]
        - mean_piston_speed_pos_pow             [m/s]
        - fuel_mep_pos_pow                      [bar]
        - fuel_consumption_pos_pow              [g/sec]
        - willans_a                             [g/kWh]
        - willans_b                             [g/h]
        - specific_fuel_consumption             [g/kWh]
        - indicated_efficiency                  [-]
        - willans_efficiency                    [-]

    :rtype: dict
    """

    factors = []

    for i, j in phases_indices:
        factors.append(calculate_willans_factors(
            params, engine_fuel_lower_heating_value, engine_stroke,
            engine_capacity, min_engine_on_speed, fmep_model,
            engine_speeds_out[i:j], engine_powers_out[i:j], times[i:j],
            velocities[i:j], accelerations[i:j], motive_powers[i:j],
            engine_temperatures[i:j], missing_powers[i:j],
            angle_slopes[i:j]
        ))

    return factors


dsp.add_data(
    'enable_phases_willans', dfl.values.enable_phases_willans,
    description='Enable the calculation of Willans coefficients for '
                'all phases?'
)

dsp.add_function(
    function=sh.add_args(calculate_phases_willans_factors),
    inputs=[
        'enable_phases_willans', 'co2_params_calibrated',
        'engine_fuel_lower_heating_value', 'engine_stroke', 'engine_capacity',
        'min_engine_on_speed', 'fmep_model', 'times', 'phases_indices',
        'engine_speeds_out', 'engine_powers_out', 'velocities', 'accelerations',
        'motive_powers', 'engine_temperatures', 'missing_powers',
        'angle_slopes'
    ],
    outputs=['phases_willans_factors'],
    input_domain=co2_utl.check_first_arg
)


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['optimal_efficiency'])
def calculate_optimal_efficiency(co2_params_calibrated, mean_piston_speeds):
    """
    Calculates the optimal efficiency [-] and t.

    :param co2_params_calibrated:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type co2_params_calibrated: lmfit.Parameters

    :param mean_piston_speeds:
        Mean piston speed vector [m/s].
    :type mean_piston_speeds: numpy.array

    :return:
        Optimal efficiency and the respective parameters:

        - mean_piston_speeds [m/s],
        - engine_bmep [bar],
        - efficiency [-].

    :rtype: dict[str | tuple]
    """
    # noinspection PyProtectedMember
    from .engine.fc import _fuel_ABC
    n_s = np.linspace(mean_piston_speeds.min(), mean_piston_speeds.max(), 10)

    A, B, C = _fuel_ABC(n_s, **co2_params_calibrated)
    # noinspection PyTypeChecker
    b = np.isclose(A, 0.0)
    # noinspection PyTypeChecker
    A = np.where(b, np.sign(C) * dfl.EPS, A)
    ac4, B2 = 4 * A * C, B ** 2
    sabc = np.sqrt(ac4 * B2)
    n = sabc - ac4

    bmep = np.where(b, np.nan, 2 * C - sabc / (2 * A))
    eff = n / (B - np.sqrt(B2 - sabc - n))

    return dict(mean_piston_speeds=n_s, engine_bmep=bmep, efficiency=eff)


@sh.add_function(dsp, outputs=['phases_fuel_consumptions'])
def calculate_phases_fuel_consumptions(
        phases_co2_emissions, fuel_carbon_content, fuel_density):
    """
    Calculates cycle phases fuel consumption [l/100km].

    :param phases_co2_emissions:
        CO2 emission of cycle phases [CO2g/km].
    :type phases_co2_emissions: numpy.array

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :param fuel_density:
        Fuel density [g/l].
    :type fuel_density: float

    :return:
        Fuel consumption of cycle phases [l/100km].
    :rtype: tuple
    """

    c = 100.0 / (fuel_density * fuel_carbon_content)

    return tuple(np.asarray(phases_co2_emissions) * c)


@sh.add_function(dsp, outputs=['fuel_consumption_value'])
def calculate_fuel_consumption_value(
        phases_fuel_consumptions, phases_distances):
    """
    Calculates the fuel consumption of the cycle [l/100km].

    :param phases_fuel_consumptions:
        Fuel consumption of cycle phases [l/100km].
    :type phases_fuel_consumptions: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array | float

    :return:
        Fuel consumption of the cycle [l/100km].
    :rtype: float
    """
    return calculate_co2_emission_value(
        phases_fuel_consumptions, phases_distances
    )


@sh.add_function(dsp, outputs=['phases_distances'])
def calculate_phases_distances(phases_indices, distances):
    """
    Calculates cycle phases distances [km].

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :param distances:
        Cumulative distance vector [m].
    :type distances: numpy.array

    :return:
        Cycle phases distances [km].
    :rtype: numpy.array
    """
    i = phases_indices.copy()
    i[:, 1] -= 1
    return np.diff(distances[i], axis=1).ravel() / 1000.0


@sh.add_function(dsp, outputs=['phases_times'])
def calculate_phases_times(phases_indices, times):
    """
    Calculates cycle phases times [s].

    :param phases_indices:
        Indices of the cycle phases [-].
    :type phases_indices: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :return:
        Cycle phases times [s].
    :rtype: numpy.array
    """
    i = phases_indices.copy()
    i[:, 1] -= 1
    return np.diff(times[i], axis=1).ravel()


@sh.add_function(dsp, outputs=['fuel_density'])
def default_fuel_density(fuel_type):
    """
    Returns the default fuel density [g/l].

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :return:
        Fuel density [g/l].
    :rtype: float
    """
    return dfl.functions.default_fuel_density.FUEL_DENSITY[fuel_type]


@sh.add_function(dsp, outputs=['engine_fuel_lower_heating_value'])
def default_engine_fuel_lower_heating_value(fuel_type):
    """
    Returns the default fuel lower heating value [kJ/kg].

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :return:
        Fuel lower heating value [kJ/kg].
    :rtype: float
    """
    return dfl.functions.default_fuel_lower_heating_value.LHV[fuel_type]


@sh.add_function(dsp, outputs=['fuel_carbon_content'], weight=3)
def default_fuel_carbon_content(fuel_type):
    """
    Returns the default fuel carbon content [CO2g/g].

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :return:
        Fuel carbon content [CO2g/g].
    :rtype: float
    """
    return dfl.functions.default_fuel_carbon_content.CARBON_CONTENT[fuel_type]


@sh.add_function(dsp, outputs=['fuel_carbon_content'])
def calculate_fuel_carbon_content(fuel_carbon_content_percentage):
    """
    Calculates the fuel carbon content as CO2g/g.

    :param fuel_carbon_content_percentage:
        Fuel carbon content [%].
    :type fuel_carbon_content_percentage: float

    :return:
        Fuel carbon content [CO2g/g].
    :rtype: float
    """
    return fuel_carbon_content_percentage / 100.0 * 44.0 / 12.0


@sh.add_function(dsp, outputs=['fuel_carbon_content_percentage'])
def calculate_fuel_carbon_content_percentage(fuel_carbon_content):
    """
    Calculates the fuel carbon content as %.

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :return:
        Fuel carbon content [%].
    :rtype: float
    """

    return fuel_carbon_content / calculate_fuel_carbon_content(1.0)


@sh.add_function(dsp, outputs=['fuel_heating_value'])
def calculate_fuel_heating_value(engine_fuel_lower_heating_value, fuel_density):
    """
    Calculates the fuel heating value as kWh/l.

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param fuel_density:
        Fuel density [g/l].
    :type fuel_density: float

    :return:
        Fuel heating value [kWh/l].
    :rtype: float
    """
    return engine_fuel_lower_heating_value * fuel_density / 36e5


@sh.add_function(dsp, outputs=['fuel_consumptions_liters_value'])
def calculate_fuel_consumptions_liters_value(times, fuel_consumptions_liters):
    """

    :param times:
    :param fuel_consumptions_liters:
    :return:
    """
    from scipy.integrate import cumtrapz
    return cumtrapz(fuel_consumptions_liters / 3600, times, initial=0)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to load data from a CO2MPAS input file.

Sub-Modules:

.. currentmodule:: co2mpas.core.load

.. autosummary::
    :nosignatures:
    :toctree: load/

    excel
    schema
    validate
"""
import io
import logging
import functools
import schedula as sh
from .excel import parse_excel_file
from .validate import dsp as _validate

try:
    from co2mpas_dice import dsp as _dice
except ImportError:
    _dice = None

log = logging.getLogger(__name__)

dsp = sh.BlueDispatcher(
    name='load_inputs',
    description='Loads from files the inputs for the CO2MPAS model.'
)


@sh.add_function(dsp, outputs=['input_file'])
def open_input_file(input_file_name):
    """
    Open the input file.

    :param input_file_name:
        Input file name.
    :type input_file_name: str

    :return:
        Input file.
    :rtype: io.BytesIO
    """
    with open(input_file_name, 'rb') as file:
        return io.BytesIO(file.read())


# noinspection PyUnusedLocal
def check_file_format(input_file_name, *args, ext=('.xlsx',)):
    """
    Check file format extension.

    :param input_file_name:
        Input file name.
    :type input_file_name: str

    :param ext:
        Allowed extensions.
    :type ext: tuple[str]

    :return:
        If the extension of the input file is within the allowed extensions.
    :rtype: bool
    """
    return input_file_name.lower().endswith(ext)


def load_from_dill(input_file):
    """
    Load inputs from .dill file.

    :param input_file:
        Input file.
    :type input_file: io.BytesIO

    :return:
        Raw input data.
    :rtype: dict
    """
    import dill
    return dill.load(input_file)


dsp.add_function(
    function=sh.add_args(load_from_dill),
    inputs=['input_file_name', 'input_file'],
    outputs=['raw_data'],
    input_domain=functools.partial(check_file_format, ext=('.dill',))
)

dsp.add_function(
    function=parse_excel_file,
    inputs=['input_file_name', 'input_file'],
    outputs=['raw_data'],
    input_domain=check_file_format
)

if _dice is not None:
    _out, _inp = ['base', 'meta', 'dice'], [
        'input_file_name', 'input_file', 'encryption_keys',
        'encryption_keys_passwords'
    ]
    # noinspection PyProtectedMember
    dsp.add_function(
        function=sh.Blueprint(sh.SubDispatchFunction(
            _dice, function_id='load_ta_file', inputs=_inp[1:], outputs=_out
        ))._set_cls(sh.add_args),
        description='Load inputs from .co2mpas.ta file.',
        inputs=_inp,
        outputs=['raw_data'],
        filters=[functools.partial(sh.map_list, [_out])],
        input_domain=functools.partial(check_file_format, ext=(
            '.co2mpas.ta', '.dice.ta', '.jet.ta'
        ))
    )

    _out, _inp = ['data', 'dice'], ['input_file_name', 'input_file']
    # noinspection PyProtectedMember
    dsp.add_function(
        function=sh.Blueprint(sh.SubDispatchFunction(
            _dice, inputs=_inp[1:], outputs=_out
        ))._set_cls(sh.add_args),
        function_id='load_co2mpas_file',
        description='Load inputs from .co2mpas file.',
        inputs=_inp,
        outputs=['raw_data'],
        filters=[functools.partial(sh.map_list, [{}, 'dice'])],
        input_domain=functools.partial(check_file_format, ext=('.co2mpas',))
    )
else:
    dsp.add_data(
        'encryption_keys_passwords', description='Encryption keys passwords.'
    )


@sh.add_function(dsp, inputs_kwargs=True, outputs=['data'])
def merge_data(
        raw_data, cmd_flags=None, hard_validation=False, declaration_mode=False,
        type_approval_mode=False, encryption_keys=None, sign_key=None,
        enable_selector=False):
    """
    Merge raw data with model flags.

    :param raw_data:
        Raw input data.
    :type raw_data: dict

    :param cmd_flags:
        Command line options.
    :type cmd_flags: dict

    :param hard_validation:
        Add extra data validations.
    :type hard_validation: bool

    :param declaration_mode:
        Use only the declaration data.
    :type declaration_mode: bool

    :param type_approval_mode:
        Is launched for TA?
    :type type_approval_mode: bool

    :param encryption_keys:
        Encryption keys for TA mode.
    :type encryption_keys: str

    :param sign_key:
        User signature key for TA mode.
    :type sign_key: str

    :param enable_selector:
        Enable the selection of the best model to predict both H/L cycles.
    :type enable_selector: bool

    :return:
        Merged raw data.
    :rtype: dict
    """
    flag = {k: v for k, v in dict(
        hard_validation=hard_validation,
        declaration_mode=declaration_mode,
        type_approval_mode=type_approval_mode,
        encryption_keys=encryption_keys,
        sign_key=sign_key,
        enable_selector=enable_selector
    ).items() if v is not None}
    data = sh.combine_dicts(raw_data, flag)
    data['flag'] = sh.combine_dicts(data.get('flag', {}), cmd_flags or {}, flag)
    return data


def check_validation(sol):
    """
    Check if the data are verified.

    :param sol:
        Validation solution.
    :type sol: schedula.Solution

    :return:
        Validated data
    :rtype: list[dict]
    """
    sol = sol.get('verified') and sol or {}
    keys = 'plan', 'flag', 'dice', 'meta', 'base'
    return [sol.get('validated_%s' % k, sh.NONE) for k in keys]


dsp.add_function(
    function=sh.SubDispatch(_validate),
    inputs=['data'],
    outputs=['plan', 'flag', 'dice', 'meta', 'base'],
    filters=[check_validation]
)



# -*- coding: utf-8 -*-
# !/usr/bin/env python
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Defines package metadata.
"""

__all__ = ['__version__', '__updated__', '__title__', '__author__',
           '__license__', '__copyright__', '__file_version__', '__uri__']

#: Authoritative project's PEP 440 version.
__version__ = version = "4.3.4"  # Also update README.rst

#: The :term:`Semantic Version` for Input/Output files.
__file_version__ = "4.4.1"

# Please UPDATE TIMESTAMP WHEN BUMPING VERSIONS AND BEFORE RELEASE.
#: Release date.
__updated__ = "2022-11-09 15:45:00"

__title__ = 'co2mpas'

__author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'

__license__ = 'EUPL, see LICENSE.txt'

__uri__ = "https://co2mpas.io"

__copyright__ = "Copyright (C) 2015-2022 European Commission (JRC)"

if __name__ == '__main__':
    import sys

    out = ';'.join(
        eval(a[2:].replace('-', '_')) for a in sys.argv[1:] if a[:2] == '--'
    )
    sys.stdout.write(out)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It contains classes and functions of general utility.
"""
import contextlib
import schedula as sh
import statistics
import itertools
import numpy as np


# noinspection PyMissingOrEmptyDocstring
class Constants(dict):
    def load(self, file, **kw):
        import yaml
        kw['Loader'] = kw.get('Loader', yaml.CLoader)
        with open(file, 'rb') as f:
            self.from_dict(yaml.load(f, **kw))
        return self

    def dump(self, file, default_flow_style=False, **kw):
        import yaml
        kw['Dumper'] = kw.get('Dumper', yaml.CDumper)
        with open(file, 'w') as f:
            yaml.dump(
                self.to_dict(), f, default_flow_style=default_flow_style, **kw
            )

    def from_dict(self, d):
        for k, v in sorted(d.items()):
            if isinstance(v, dict) and '__constants__' in v:
                o = getattr(self, k, Constants())
                if isinstance(o, Constants):
                    v = o.from_dict(v['__constants__'])
                elif issubclass(o.__class__, Constants) or \
                        issubclass(o, Constants):
                    v = o().from_dict(v['__constants__'])
                if not v:
                    continue
            elif hasattr(self, k) and getattr(self, k) == v:
                continue
            setattr(self, k, v)
            self[k] = v

        return self

    def to_dict(self):
        import inspect
        s, pr = set(dir(self)) - set(dir(Constants)), {}
        for n in s.union(self.__class__.__dict__.keys()):
            if n.startswith('__'):
                continue
            v = getattr(self, n)
            if inspect.ismethod(v) or inspect.isbuiltin(v):
                continue
            if isinstance(v, Constants):
                pr[n] = {'__constants__': v.to_dict()}
            elif inspect.isclass(v) and issubclass(v, Constants):
                # noinspection PyCallByClass,PyTypeChecker
                pr[n] = {'__constants__': v.to_dict(v)}
            else:
                pr[n] = v
        return pr


# noinspection PyMissingOrEmptyDocstring
class List(list):
    empty = sh.EMPTY
    dtype = None

    def __new__(cls, *args, dtype=float, **kwargs):
        obj = super(List, cls).__new__(cls, *args, **kwargs)
        obj.dtype = dtype
        return obj

    def __getitem__(self, item):
        r = super(List, self).__getitem__(item)
        if r is self.empty:
            raise IndexError('list index out of range')
        elif isinstance(item, slice):
            return self.__class__(r)
        return r

    def __setitem__(self, key, value):
        try:
            return super(List, self).__setitem__(key, value)
        except IndexError:
            self.extend([self.empty] * (key - len(self)))
            self.append(value)
            return super(List, self).__setitem__(key, value)

    def toarray(self, dtype=None, *args, **kwargs):
        return np.array(self, dtype or self.dtype, *args, **kwargs)


@contextlib.contextmanager
def numpy_random_seed(seed):
    """
    Set temporary the numpy random state.

    :param seed:
        Seed for `RandomState`.
    :type seed: int
    """
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def argmax(values, **kws):
    """
    Returns the indices of the maximum values along an axis.

    :param values:
        Input array.
    :type values: numpy.array | list

    :return:
        Indices of the maximum values
    :rtype: numpy.ndarray
    """
    return np.argmax(np.append(values, [True]), **kws)


def mae(x, y, w=None):
    """
    Mean absolute error.

    :param x:
        Reference values.
    :type x: numpy.array

    :param y:
        Output values.
    :type y: numpy.array

    :param w:
        Weights.
    :type w: numpy.array

    :return:
        Mean absolute error.
    :rtype: float
    """
    if w is not None:
        return (np.abs(x - y) * w).sum() / w.sum()
    return np.mean(np.abs(x - y))


def mad(x, med=None):
    """
    Median Absolute Deviation.
    """
    med = np.nanmedian(x) if med is None else med
    return np.nanmedian(np.abs(x - med))


def sliding_window(xy, dx_window):
    """
    Returns a sliding window (of width dx) over data from the iterable.

    :param xy:
        X and Y values.
    :type xy: list[(float, float) | list[float]]

    :param dx_window:
        dX window.
    :type dx_window: float

    :return:
        Data (x & y) inside the time window.
    :rtype: generator
    """

    dx = dx_window / 2
    it = iter(xy)
    v = next(it)
    window = []

    for x, y in xy:
        # window limits
        x_dn = x - dx
        x_up = x + dx

        # remove samples
        window = [w for w in window if w[0] >= x_dn]

        # add samples
        while v and v[0] <= x_up:
            window.append(v)
            try:
                v = next(it)
            except StopIteration:
                v = None

        yield window


# noinspection PyShadowingBuiltins
def median_filter(x, y, dx_window, filter=statistics.median_high):
    """
    Calculates the moving median-high of y values over a constant dx.

    :param x:
        x data.
    :type x: Iterable

    :param y:
        y data.
    :type y: Iterable

    :param dx_window:
        dx window.
    :type dx_window: float

    :param filter:
        Filter function.
    :type filter: callable

    :return:
        Moving median-high of y values over a constant dx.
    :rtype: numpy.array
    """

    xy = list(zip(x, y))
    _y = []
    add = _y.append
    for v in sliding_window(xy, dx_window):
        add(filter(list(zip(*v))[1]))
    return np.array(_y)


def get_inliers(x, n=1, med=np.median, std=np.std):
    """
    Returns the inliers data.

    :param x:
        Input data.
    :type x: Iterable

    :param n:
        Number of standard deviations.
    :type n: int

    :param med:
        Median function.
    :type med: callable, optional

    :param std:
        Standard deviation function.
    :type std: callable, optional

    :return:
         Inliers mask, median and standard deviation of inliers.
    :rtype: (numpy.array, float, float)
    """
    x = np.asarray(x)
    if not x.size:
        return np.zeros_like(x, dtype=bool), np.nan, np.nan
    m, s = med(x), std(x)
    with np.errstate(divide='ignore', invalid='ignore'):
        y = n > (np.abs(x - m) / s)
    return y, m, s


def reject_outliers(x, n=1, med=np.median, std=np.std):
    """
    Calculates the median and standard deviation of the sample rejecting the
    outliers.

    :param x:
        Input data.
    :type x: Iterable

    :param n:
        Number of standard deviations.
    :type n: int

    :param med:
        Median function.
    :type med: callable, optional

    :param std:
        Standard deviation function.
    :type std: callable, optional

    :return:
        Median and standard deviation.
    :rtype: (float, float)
    """

    y, m, s = get_inliers(x, n=n, med=med, std=std)

    if y.any():
        y = np.asarray(x)[y]

        m, s = med(y), std(y)

    return m, s


def clear_fluctuations(times, gears, dt_window):
    """
    Clears the gear identification fluctuations.

    :param times:
        Time vector.
    :type times: numpy.array

    :param gears:
        Gear vector.
    :type gears: numpy.array

    :param dt_window:
        Time window.
    :type dt_window: float

    :return:
        Gear vector corrected from fluctuations.
    :rtype: numpy.array
    """

    xy = [list(v) for v in zip(times, gears)]

    for samples in sliding_window(xy, dt_window):

        up, dn = False, False

        x, y = zip(*samples)

        for k, d in enumerate(np.diff(y)):
            if d > 0:
                up = True
            elif d < 0:
                dn = True

            if up and dn:
                m = statistics.median_high(y)
                for v in samples:
                    v[1] = m
                break

    return np.array([y[1] for y in xy])


# noinspection PyUnusedLocal
def check_first_arg(first, *args):
    """
    Check first arg is true.

    :param first:
        First arg.
    :type first: T

    :return:
        If first arg is true.
    :rtype: bool
    """
    return bool(first)


# noinspection PyUnusedLocal
def check_first_arg_false(first, *args):
    """
    Check first arg is false.

    :param first:
        First arg.
    :type first: T

    :return:
        If first arg is false.
    :rtype: bool
    """
    return not bool(first)


def index_phases(phases):
    """
    Return the indices of the phases when is true.

    :param phases:
        Phases vector.
    :type phases: numpy.array

    :return:
        Indices of the phases when is true.
    :rtype: numpy.array
    """
    i = np.where(np.logical_xor(phases[:-1], phases[1:]))[0] + 1
    if i.shape[0]:
        if i[0] and phases[0]:
            i = np.append([0], i)
        if phases[-1]:
            i = np.append(i, [len(phases) - 1])
    elif phases[0]:
        i = np.append([0, len(phases) - 1], i)
    return i.reshape(-1, 2)


def pairwise(iterable):
    """
    A sequence of overlapping sub-sequences.

    :param iterable:
        An iterable object.
    :type iterable: iterable

    :return:
        A zip object.
    :rtype: zip

    Example::

        >>> list(pairwise([1, 2, 3, 4, 5]))
        [(1, 2), (2, 3), (3, 4), (4, 5)]
    """

    a, b = itertools.tee(iterable)

    next(b, None)

    return zip(a, b)
